package FACS.Annotator;

import javax.swing.*;
import java.awt.event.*;
import javax.swing.event.*;
import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Color;
import java.awt.GridLayout;
import java.awt.Toolkit;
import java.io.*;
import java.net.URL;
import java.net.URLConnection;
import java.net.MalformedURLException;
import java.net.URLEncoder;
import java.util.*;
import java.awt.dnd.*;
import java.awt.datatransfer.*;
import javax.swing.table.*;

/**
 * Client side (applet) of web-interface experiment annotator
 * @author Mark
 *
 */
public class AnnotatorClient extends JApplet implements ActionListener, ListSelectionListener, ChangeListener {

	public static String SERVLET_URL;


	Vector files; //filenames (==annos.keys())
	Hashtable annos; //actual annotations (hashtable of annotation hashtables)
	
	FileRenderer fileRenderer;

	JList lsPlates, lsFiles;
	JTabbedPane tabPlates;
	JTextField tfCell;
	JTable plate; JScrollPane spPlate; JList rowHeader;
	JComboBox cbProp; boolean isInternalChange; //for changing cb without changing props selection
	JComboBox cbSort;
	JTable props;

	boolean isPlateFitToScreen;


	public void init() {

		SERVLET_URL = "servlets/AnnotatorServlet";
		if(getParameter("servlet") != null)
			SERVLET_URL = getParameter("servlet");

		establishConnection();
		createGUI();
    }

	public void establishConnection() { //talk to server, get files/regions
		try {
			URL servletURL = new URL(getCodeBase(),SERVLET_URL);
			URLConnection servlet = servletURL.openConnection();
			servlet.setDoInput(true);
			servlet.setDoOutput(true);

			servlet.setUseCaches (false); //don't cache connection
			servlet.setDefaultUseCaches (false);

			PrintStream out = new PrintStream(servlet.getOutputStream());
			out.print("directory="+getParameter("directory"));
			out.print("&mode="+"get");
			out.close();

			//get files
			ObjectInputStream ois = new ObjectInputStream(new BufferedInputStream(servlet.getInputStream()));
			try {
				annos = (Hashtable)ois.readObject();
			} catch (ClassNotFoundException cnfe) {
				cnfe.printStackTrace();
				annos = new Hashtable();
			}

			files = new Vector(annos.keySet());
			Collections.sort(files); //put filenames in alphabetical order

			lsFiles = new JList();
			fileRenderer = new FileRenderer(annos);
			lsFiles.setModel(new DefaultListModel());
			lsFiles.setCellRenderer(fileRenderer);
			for(int f = 0; f < files.size(); f++) {
				((DefaultListModel)lsFiles.getModel()).addElement(files.elementAt(f));
			}

			//get plates
			lsPlates = new JList();
			tabPlates = new JTabbedPane();
			lsPlates.setModel(new DefaultListModel());
			lsPlates.setCellRenderer(new PlateRenderer());
			Vector plates = new Vector();
			try {
				plates = (Vector)ois.readObject();
			} catch (ClassNotFoundException cnfe) {
				cnfe.printStackTrace();
			}
			if(plates.size() == 0) {
				Plate p1 = new Plate((JList)lsPlates,tabPlates,"Controls",1);
				((DefaultListModel)lsPlates.getModel()).addElement(p1);
				tabPlates.addTab(p1.getFullName(),new NullComponent());
				Plate p2 = new Plate((JList)lsPlates,tabPlates,1);
				((DefaultListModel)lsPlates.getModel()).addElement(p2);
				tabPlates.addTab(p2.getFullName(),new NullComponent());				
				
				/*((DefaultListModel)lsPlates.getModel()).addElement(new Plate((PlateList)lsPlates,"Controls",1));
				((DefaultListModel)lsPlates.getModel()).addElement(new Plate((PlateList)lsPlates,1));*/
			}
			for(int i = 0; i < plates.size(); i++) {
				Plate p = (Plate)plates.elementAt(i);
				p.setListenerList(); //for some reason, this doesn't get initialized (it is transient, of course)
				p.pl = (JList)lsPlates; //this also won't get initialized
				//need to place FACSData [files] into here (& remove from lsFiles)
				replaceFDs(p);
				((DefaultListModel)lsPlates.getModel()).addElement(p);
				tabPlates.addTab(p.getFullName(),new NullComponent());
			}
			ois.close();

		} catch (java.net.UnknownHostException e) {
            System.err.println(e.toString());
        } catch (IOException ioe) {
			ioe.printStackTrace();
		}
        tabPlates.setPreferredSize(new Dimension(100,30));
	}

	public void replaceFDs(Plate p) {
		for(int f = 0; f < p.getData().size(); f++) {
			String fname = (String)p.getData().elementAt(f);
			if(!files.contains(fname)) {
				p.getData().setElementAt(new Hashtable(),f);
				continue;
			}					
			p.getData().setElementAt(annos.get(fname),f);
			((DefaultListModel)lsFiles.getModel()).removeElement(fname);
		}
	}

	public void createGUI() {
		setBackground(Color.white);
		JPanel p = new JPanel();
		p.setBackground(Color.white);

		p.setLayout(new BorderLayout());

		JPanel pRight = new JPanel();
		
		//Plates
		JPanel pWest = new JPanel();
		pWest.setLayout(new BoxLayout(pWest,BoxLayout.Y_AXIS));
		JPanel pButs = new JPanel();
		pButs.setLayout(new GridLayout(1,2));
		JButton bNewP = new JButton("New Plate");
		bNewP.setActionCommand("newp");
		bNewP.addActionListener(this);
		pButs.add(bNewP);
		JButton bDelP = new JButton("Remove Plate");
		bDelP.setActionCommand("delp");
		bDelP.addActionListener(this);
		pButs.add(bDelP);
		pWest.add(pButs);

		
		lsPlates.setVisibleRowCount(-1);
		lsPlates.addListSelectionListener(this);
		lsPlates.addMouseListener(new MouseAdapter() { //double click -> change row/col of Plate
			public void mouseClicked(MouseEvent me) {
				if(me.getClickCount() > 1) {
					int i = lsPlates.locationToIndex(me.getPoint());
					Plate p = (Plate)lsPlates.getModel().getElementAt(i);
					p.resizePlate();
				}
			}
		});
		lsPlates.addKeyListener(new KeyAdapter() {
			public void keyReleased(KeyEvent ke) {
				if(ke.getKeyCode() == KeyEvent.VK_DELETE || ke.getKeyCode() == KeyEvent.VK_BACK_SPACE) {
					deleteSelectedPlate();
					ke.consume();
				}
			}
		});
		JScrollPane spPlates = new JScrollPane(lsPlates);
		spPlates.setPreferredSize(new Dimension(190,200));
		pWest.add(spPlates);
		pWest.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(),"Plates"));
		
		JPanel pMButs = new JPanel();
		pMButs.setLayout(new GridLayout(1,2));
		JButton bMvUp = new JButton("Move Up");
		bMvUp.setActionCommand("mv_up");
		bMvUp.addActionListener(this);
		pMButs.add(bMvUp);
		JButton bMvDn = new JButton("Move Down");
		bMvDn.setActionCommand("mv_dn");
		bMvDn.addActionListener(this);
		pMButs.add(bMvDn);
		pWest.add(pMButs);
		
		pRight.add(pWest);
		
		//Arrow buttons to move files
		JPanel pArrows = new JPanel();
		pArrows.setLayout(new GridLayout(2,1));
		JButton bAdd = new JButton ("<-");
		bAdd.setActionCommand("add_file");
		bAdd.addActionListener(this);
		pArrows.add(bAdd);
		JButton bRem = new JButton ("->");
		bRem.setActionCommand("rem_file");
		bRem.addActionListener(this);
		pArrows.add(bRem);
		pRight.add(pArrows);

		//Right pane (files)
		//JTabbedPane pEast = new JTabbedPane();
		JPanel pEast = new JPanel();
		//Data files
		JPanel pFiles = new JPanel(); //if I want to sort them...
		pFiles.setLayout(new BoxLayout(pFiles,BoxLayout.Y_AXIS));
		
		pFiles.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(),"Unassigned Files"));
		JPanel pSort = new JPanel();
		cbSort = new JComboBox();
		cbSort.addActionListener(this);
		cbSort.setActionCommand("files_anno");
		pSort.add(cbSort);
		JButton bSort = new JButton("Sort");
		bSort.setActionCommand("sort");
		bSort.addActionListener(this);
		pSort.add(bSort);
		pFiles.add(pSort);
		lsFiles.setVisibleRowCount(-1);
		/*lsFiles.setDragEnabled(true);
		lsFiles.setTransferHandler(new ListTransferHandler());*/
		JScrollPane spFiles = new JScrollPane(lsFiles);
		spFiles.setPreferredSize(new Dimension(370,190));
		pFiles.add(spFiles);
		pEast.add(pFiles);
		pRight.add(pEast);
		p.add(pRight, BorderLayout.SOUTH);

		//JTable that displays plate and tabs
		JPanel pPlate = new JPanel();
		pPlate.setLayout(new BorderLayout());
		
		//Tabs for plates (in addition to JList)
		pPlate.add(tabPlates,BorderLayout.NORTH);
		tabPlates.addChangeListener(this);
		
		//Top panel (buttons, cell value, variable drop-down)
		JPanel pPTop = new JPanel();
		
		JButton bOk = new JButton("Save Plates");
		bOk.setActionCommand("okay");
		bOk.addActionListener(this);
		pPTop.add(bOk);

		JPanel pCB = new JPanel();
		pCB.setLayout(new GridLayout(2,1));
		cbProp = new JComboBox();
		cbProp.setEditable(true);
		cbProp.addActionListener(this);
		cbProp.setActionCommand("prop");
		populateCB();
		pCB.add(cbProp);
		pPTop.add(pCB);
		

		JPanel pTFCell = new JPanel();
		pTFCell.setLayout(new GridLayout(2,1));
		tfCell = new JTextField(40);
		tfCell.addKeyListener(new KeyAdapter() {
			public void keyReleased(KeyEvent ke) {
				if(ke.getKeyCode() == KeyEvent.VK_ENTER)
					setCellValue();
			}
		});
		pTFCell.add(tfCell);
		pTFCell.add(new JLabel(""));
		pPTop.add(pTFCell);
		
		JPanel pTopRight = new JPanel();
		pTopRight.setLayout(new GridLayout(2,1));
		JButton pTrans = new JButton("Transpose Plate");
		pTrans.setActionCommand("trans");
		pTrans.addActionListener(this);
		pTopRight.add(pTrans);
		JButton bBlank = new JButton("Insert Space");
		bBlank.setActionCommand("blank");
		bBlank.addActionListener(this);
		pTopRight.add(bBlank);
		pPTop.add(pTopRight);
		JButton bFit = new JButton("Fit to Screen");
		bFit.setActionCommand("fit");
		bFit.addActionListener(this);
		pPTop.add(bFit);
		p.add(pPTop,BorderLayout.NORTH);
		
		//Plate
		plate = new JTable();
		plate.setCellSelectionEnabled(true);
		plate.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);
		isPlateFitToScreen = false;
		plate.getTableHeader().setReorderingAllowed(false);
		//column selection
		plate.getTableHeader().addMouseListener(new MouseAdapter() {
			boolean mousedown = false;
			public void mousePressed(MouseEvent me) {
				if(!mousedown && !(me.isControlDown())) { //start new selection
					clearTableSelection();
				}
				mousedown = true;
				JTableHeader th = (JTableHeader)me.getSource();
				if(me.isControlDown())
					selectColOR(th.columnAtPoint(me.getPoint()));
				else
					selectCol(th.columnAtPoint(me.getPoint()));
			}
			public void mouseReleased(MouseEvent me) {
				mousedown = false;
			}
		});
		plate.getTableHeader().addMouseMotionListener(new MouseMotionAdapter() {
			public void mouseDragged(MouseEvent me) {
				JTableHeader th = (JTableHeader)me.getSource();
				selectCol(th.columnAtPoint(me.getPoint()));
			}
		});
		//row selection
		rowHeader = new JList();
		rowHeader.addMouseListener(new MouseAdapter() {
			boolean mousedown = false;
			public void mousePressed(MouseEvent me) {
				if(!mousedown && !(me.isControlDown())) { //start new selection
					clearTableSelection();
				}
				mousedown = true;
				JList th = (JList)me.getSource();
				if(me.isControlDown())
					selectRowOR(th.locationToIndex(me.getPoint()));
				else
					selectRow(th.locationToIndex(me.getPoint()));
			}
			public void mouseReleased(MouseEvent me) {
				mousedown = false;
			}
		});
		rowHeader.addMouseMotionListener(new MouseMotionAdapter() {
			public void mouseDragged(MouseEvent me) {
				JList th = (JList)me.getSource();
				selectRow(th.locationToIndex(me.getPoint()));
			}
		});
		//cell selection (i.e. if click inside table)
		plate.addMouseListener(new MouseAdapter() {
			/*public void mousePressed(MouseEvent me) {
				setCellSelectable();
			}*/
			public void mouseReleased(MouseEvent me) { //display cell contents in top textfield
				setCellSelectable();
				displayCell(me.getPoint());
			}
		});
		//delete action
		plate.addKeyListener(new KeyAdapter() {
			public void keyReleased(KeyEvent ke) {
				if(ke.getKeyCode() == KeyEvent.VK_DELETE || ke.getKeyCode() == KeyEvent.VK_BACK_SPACE)
					deleteCellValue(false);
			}
		});
		//scroll pane for table
		spPlate = new JScrollPane(plate);
		pPlate.add(spPlate, BorderLayout.CENTER);
		//p.add(pPlate);
		
		//Properties
		JPanel pProps = new JPanel();
		props = new JTable(new PropertiesTableModel(this));
		props.addMouseListener(new MouseAdapter() { //edit with one click
			public void mouseClicked(MouseEvent me) {
				JTable table = (JTable)me.getSource();
				int row = table.rowAtPoint(me.getPoint());
				int column = table.columnAtPoint(me.getPoint());
				if(column == 1) {
					table.editCellAt(row, column);
					table.setEditingRow(row);
					table.setEditingColumn(column);
				} else {
					table.changeSelection(row,0,me.isControlDown(),true);
				}
				updateCBSel(row); //update plate key, cb selection
				table.repaint();
			}
		});
		//transfer copy events to plate
		props.setTransferHandler(null);
		props.addKeyListener(new KeyAdapter() {
			public void keyReleased(KeyEvent ke) {
				if(ke.getKeyCode() == KeyEvent.VK_C && ke.isControlDown()) {
					transferCopy();
					ke.consume();
				}
			}
		});

		JScrollPane spProps = new JScrollPane(props);
		spProps.setPreferredSize(new Dimension(200,180));
		spProps.setColumnHeader(null);
		
		spProps.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(),"Selected Cell:"));
		pProps.add(spProps);
		pPlate.add(pProps, BorderLayout.EAST);
		p.add(pPlate, BorderLayout.CENTER);

		/*pRight.add(pEast);
		p.add(pRight, BorderLayout.EAST);
		
		JButton bOk = new JButton("OK");
		bOk.setActionCommand("okay");
		bOk.addActionListener(this);
		p.add(bOk, BorderLayout.SOUTH);*/

		getContentPane().add(p);

		getContentPane().validate();
		//spPlate.setPreferredSize(new Dimension(getContentPane().getWidth()-pEast.getWidth()-pWest.getWidth(),500));
		spPlate.setPreferredSize(new Dimension(500,160));
		getContentPane().validate();
	}
	
	public void transferCopy() {//transfer copy command to plate
        //plate.dispatchEvent(new KeyEvent(plate, KeyEvent.KEY_RELEASED, System.currentTimeMillis(),KeyEvent.CTRL_DOWN_MASK, KeyEvent.VK_C, 'c'));
		
		//implementing this feature requires a signed applet.
		plate.getTransferHandler().exportToClipboard(plate,this.getToolkit().getSystemClipboard(),TransferHandler.COPY);
	}
	
	public void deleteSelectedPlate() {
		Plate p = (Plate)lsPlates.getSelectedValue();
		int i = lsPlates.getSelectedIndex();
		if(p.getData().size() == 0) {
			((DefaultListModel)lsPlates.getModel()).remove(i);
			tabPlates.removeTabAt(i);
			lsPlates.setSelectedIndex(0);			
		} else {
			JOptionPane.showMessageDialog(null,"Error: Plate is not Empty!");
		}
	}

	//functions for handling table row/col selection
	public void selectCol(int col) { //click on column header -> select that column
		plate.setColumnSelectionAllowed(true);
		plate.setRowSelectionAllowed(false);
		plate.addColumnSelectionInterval(col,col);
	}

	public void selectColOR(int col) { //will DEselect already selected columns
		plate.setColumnSelectionAllowed(true);
		plate.setRowSelectionAllowed(false);
		if(plate.isColumnSelected(col))
			plate.removeColumnSelectionInterval(col,col);
		else
			plate.addColumnSelectionInterval(col,col);
	}

	public void selectRow(int row) {
		plate.setRowSelectionAllowed(true);
		plate.setColumnSelectionAllowed(false);
		plate.addRowSelectionInterval(row,row);
	}

	public void selectRowOR(int row) { //will DEselect already selected rows
		plate.setRowSelectionAllowed(true);
		plate.setColumnSelectionAllowed(false);
		if(plate.isRowSelected(row))
			plate.removeRowSelectionInterval(row,row);
		else
			plate.addRowSelectionInterval(row,row);
	}

	public void clearTableSelection() {
		plate.clearSelection();
	}

	public void setCellSelectable() { //switch to cell-selection mode in table
		plate.setCellSelectionEnabled(true);
	}

	public void displayCell(java.awt.Point p) { //in textfield
		Vector sel = saveSelection(props);
		int row = plate.rowAtPoint(p);
		int col = plate.columnAtPoint(p);
		tfCell.setText((String)((PlateTableModel)plate.getModel()).getValueAt(row,col));
		tfCell.setEditable(((PlateTableModel)plate.getModel()).isCellEditable(row,col));
		Hashtable fd = ((PlateTableModel)plate.getModel()).getFDAt(row,col);
		((PropertiesTableModel)props.getModel()).setFD(fd);
		//select a row of props, if none selected (i.e. a new property that wasn't there before, but should be now...)
		int r = ((PropertiesTableModel)props.getModel()).getRow((String)cbProp.getSelectedItem());
		restoreSelection(sel,props);
		if(props.getSelectedRows().length == 0) {
			props.changeSelection(r,0,false,false);
			((PlateTableModel)plate.getModel()).setKeys(((PropertiesTableModel)props.getModel()).getKeys(props.getSelectedRows()));
		}
	}

	public Vector saveSelection(JTable table) {
		Vector out = new Vector();
		int[] rows = table.getSelectedRows();
		int[] cols = table.getSelectedColumns();
		if(cols.length == 0 && table.getRowSelectionAllowed()) {
			cols = TableTransferHandler.incrementArray(table.getColumnCount());
		}
		if(rows.length == 0 && table.getColumnSelectionAllowed()) {
			rows = TableTransferHandler.incrementArray(table.getRowCount());
		}
		out.add(rows);
		out.add(cols);
		return out;
	}

	public void restoreSelection(Vector out, JTable table) {
		table.clearSelection();
		int[] rows = (int[])out.elementAt(0);
		int[] cols = (int[])out.elementAt(1);
		for(int i = 0; i < rows.length; i++) { //rows
			table.addRowSelectionInterval(rows[i],rows[i]);
		}
		for(int i = 0; i < cols.length; i++) { //cols
			table.addColumnSelectionInterval(cols[i],cols[i]);
		}
	}

	/**
	 * When user hits "ENTER" in textfield: set all selected cells
	 * to that value (for the selected annotation).
	 */
	public void setCellValue() {
		PlateTableModel ptm = (PlateTableModel)plate.getModel();
		int[] rows = plate.getSelectedRows();
		int[] cols = plate.getSelectedColumns();

		//fix rows or cols if row-only or col-only mode (thus, value given as 0)
		if(ptm.isFillAcross()) { //can select rows
			if(cols.length == 0) {
				cols = TableTransferHandler.incrementArray(plate.getColumnCount());
			}
		} else {
			if(rows.length == 0) {
				rows = TableTransferHandler.incrementArray(plate.getRowCount());
			}
		}
		
		for(int i = 0; i < rows.length; i++) {
			for(int j = 0; j < cols.length; j++) {
				ptm.setValueAt(tfCell.getText(),rows[i],cols[j]);
			}
		}
		
		//update props table
		Hashtable fd = ((PlateTableModel)plate.getModel()).getFDAt(rows[rows.length-1],cols[cols.length-1]);
		((PropertiesTableModel)props.getModel()).setFD(fd);
	}

	public void updateCBSel(int row) { //update plate to display currently selected row in props
		isInternalChange = true;
		cbProp.setSelectedItem(props.getValueAt(row,0));
		isInternalChange = false;
	}

	public void deleteCellValue(boolean removeFile) { //delete/backspace key
		PlateTableModel model = (PlateTableModel)plate.getModel();

		int[] rows = plate.getSelectedRows();
		int[] cols = plate.getSelectedColumns();

		if(model.isFillAcross()) { //can select rows
			if(cols.length == 0) {
				cols = TableTransferHandler.incrementArray(plate.getColumnCount());
			}
		} else {
			if(rows.length == 0) {
				rows = TableTransferHandler.incrementArray(plate.getRowCount());
			}
		}

		if(model.isFileMode() || removeFile) {
			Vector fds = model.remove(rows,cols,0,0);
			for(int i = 0; i < fds.size(); i++) {
				((DefaultListModel)lsFiles.getModel()).addElement(((Hashtable)fds.elementAt(i)).get("filename"));
			}
		} else {
			for(int i = 0; i < rows.length; i++) {
				for(int j = 0; j < cols.length; j++) {
					model.setValueAt("",rows[i],cols[j]);
				}
			}
		}
	}

	public void populateCB() {
		//add all possible annotations to this combobox
		Vector s = new Vector(); Set j;
		for(int i = 0; i < files.size(); i++) {
			if(files.elementAt(i) == null)
				continue;
			j = ((Hashtable)annos.get(files.elementAt(i))).keySet();
			Iterator ij = j.iterator();
			while(ij.hasNext()) {
				Object o = ij.next();
				if(!s.contains(o))
					s.addElement(o);
			}
		}
		cbProp.removeAllItems();
		cbSort.removeAllItems();
		Collections.sort(s);
		Iterator it = s.iterator();
		while(it.hasNext()) {
			Object i = it.next();
			cbProp.addItem(i);
			cbSort.addItem(i);
		}
		cbProp.setSelectedItem("filename");
		cbSort.setSelectedItem("filename");
	}

	public void actionPerformed(ActionEvent e) {
		if(e.getActionCommand().equals("trans")) { //transpose plate
			transposePlate();
		}
		if(e.getActionCommand().equals("add_file")) { //add file(s) to a plate
			Plate p = (Plate)lsPlates.getSelectedValue();
			Object[] sel_files = lsFiles.getSelectedValues();
			LinkedList ll = TableTransferHandler.getFDsFromName(annos,sel_files);
			p.add(ll);
			
			for(int i = 0; i < sel_files.length; i++)
				((DefaultListModel)lsFiles.getModel()).removeElement(sel_files[i]);
		}
		if(e.getActionCommand().equals("rem_file")) { //remove file(s) from a plate
			deleteCellValue(true);
		}
		if(e.getActionCommand().equals("prop")) {
			try {
				Vector sel = saveSelection(plate);
				((PlateTableModel)plate.getModel()).setKey((String)cbProp.getSelectedItem());
				boolean newProp = false;
				if(((DefaultComboBoxModel)cbProp.getModel()).getIndexOf(cbProp.getSelectedItem()) == -1) {
					cbProp.addItem(cbProp.getSelectedItem());
					newProp = true;
				}
				int r = ((PropertiesTableModel)props.getModel()).getRow((String)cbProp.getSelectedItem());
				if(!isInternalChange) {
					props.changeSelection(r,0,false,false);
					if(newProp) {
						props.clearSelection();
						Hashtable h = ((PropertiesTableModel)props.getModel()).getFD();
						h.put((String)cbProp.getSelectedItem(),"");
						((PropertiesTableModel)props.getModel()).setFD(h);
						props.changeSelection(((PropertiesTableModel)props.getModel()).getRow((String)cbProp.getSelectedItem()),0,false,false);
					}
				}
				((PlateTableModel)plate.getModel()).setKeys(((PropertiesTableModel)props.getModel()).getKeys(props.getSelectedRows()));
				restoreSelection(sel,plate);
				
				tfCell.setText((String)((PlateTableModel)plate.getModel()).getValueAt(plate.getSelectedRow(),plate.getSelectedColumn()));
				tfCell.setEditable(((PlateTableModel)plate.getModel()).isCellEditable(plate.getSelectedRow(),plate.getSelectedColumn()));
			} catch(NullPointerException npe) {
			} catch(ClassCastException cce) {}
			return;
		}
		if(e.getActionCommand().equals("files_anno")) {
			//display the selected annotation for the files in lsFiles
			//don't necessarily sort
			fileRenderer.setKey((String)cbSort.getSelectedItem());
			lsFiles.repaint();
		}
		if(e.getActionCommand().equals("sort")) { //sort file list by selected anno.
			DefaultListModel lm = (DefaultListModel)lsFiles.getModel();
			Object[] list = lm.toArray();
			Arrays.sort(list,new FileComparator(annos,(String)cbSort.getSelectedItem()));
			lm.removeAllElements();
			for(int i = 0; i < list.length; i++) {
				lm.addElement(list[i]);
			}
		}
		if(e.getActionCommand().equals("okay")) { //save & exit
			try {
				save_plates();
				getAppletContext().showDocument(new URL(getCodeBase(),getParameter("away")));
			} catch (MalformedURLException murle) {
				System.out.println(murle.toString());
				stop();
			}
		}
		if(e.getActionCommand().equals("blank")) { //insert a blank into a plate
			if(plate.getCellSelectionEnabled()) { //only insert if able to select single cell
				((PlateTableModel)plate.getModel()).insert(new Hashtable(),plate.getSelectedRow(),plate.getSelectedColumn());
			}
		}
		if(e.getActionCommand().equals("newp")) { //make a new plate
			PlateDialog pd = new PlateDialog(8,12,"Plate");
			//check that there isn't a plate of the same name, else create a new one with incremented #
			Enumeration plates = ((DefaultListModel)lsPlates.getModel()).elements();
			Plate p;
			int i = 1;
			while(plates.hasMoreElements()) {
				p = (Plate)plates.nextElement();
				if(p.getName().equals(pd.name)) {
					int p_num = p.getNum();
					if(p_num >= i) {
						i = p_num+1;
					}
				}
			}
			Plate newP = new Plate((JList)lsPlates, tabPlates, pd.name, i);
			newP.setSize(pd.rows,pd.cols);
			((DefaultListModel)lsPlates.getModel()).addElement(newP);
			tabPlates.addTab(newP.getFullName(),new NullComponent());
			tabPlates.setPreferredSize(new Dimension(100,30));
		}
		if(e.getActionCommand().equals("delp")) {
			deleteSelectedPlate();
		}
		if(e.getActionCommand().equals("mv_up")) { //move a plate up in the list
			DefaultListModel dlm = (DefaultListModel)lsPlates.getModel();
			int i = lsPlates.getSelectedIndex();
			if(i == 0) //at top already
				return;
			Plate p = (Plate)dlm.remove(i);
			dlm.insertElementAt(p,i-1);
			tabPlates.remove(i);
			tabPlates.insertTab(p.getFullName(),null,new NullComponent(),null,i-1);
			lsPlates.setSelectedIndex(i-1);
		}
		if(e.getActionCommand().equals("mv_dn")) { //move a plate down in the list
			DefaultListModel dlm = (DefaultListModel)lsPlates.getModel();
			int i = lsPlates.getSelectedIndex();
			if(i == dlm.size()-1) //at bottom already
				return;
			Plate p = (Plate)dlm.remove(i);
			dlm.insertElementAt(p,i+1);
			tabPlates.remove(i);
			tabPlates.insertTab(p.getFullName(),null,new NullComponent(),null,i+1);
			lsPlates.setSelectedIndex(i+1);
		}
		if(e.getActionCommand().equals("fit")) { //fit plate to screen
			if(!isPlateFitToScreen) {
				plate.setAutoResizeMode(JTable.AUTO_RESIZE_ALL_COLUMNS);
				plate.doLayout();
				isPlateFitToScreen = true;
			} else {
				plate.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);
				plate.doLayout();
				isPlateFitToScreen = false;
			}
		}
	}

	public void valueChanged(ListSelectionEvent lse) {
		if(lse.getValueIsAdjusting())
            return;

		if(lsPlates.getSelectedValue() == null)
			return;
		
		tabPlates.setSelectedIndex(lsPlates.getSelectedIndex());
		
		PlateTableModel ptm = new PlateTableModel((Plate)lsPlates.getSelectedValue());
		AlphabetListModel lmRowH = new AlphabetListModel(ptm.getRowCount());
		ptm.addTableModelListener(lmRowH);
		plate.setModel(ptm);
		((PlateTableModel)plate.getModel()).setKey((String)(cbProp).getSelectedItem());
		((PlateTableModel)plate.getModel()).setKeys(((PropertiesTableModel)props.getModel()).getKeys(props.getSelectedRows()));
		//row header
		rowHeader.setModel(lmRowH);    
		rowHeader.setFixedCellWidth(50);
		rowHeader.setFixedCellHeight(plate.getRowHeight()); //+plate.getRowMargin()
		rowHeader.setCellRenderer(new RowHeaderRenderer(plate));
		spPlate.setRowHeaderView(rowHeader);
		//"transposition" operation (i.e. switch from fill-across to fill-down
		KeyStroke ks = KeyStroke.getKeyStroke(KeyEvent.VK_T, InputEvent.CTRL_MASK, false);
		Action aTrans = new AbstractAction() {
			public void actionPerformed(ActionEvent e) {
				transposePlate();
			}
		};
		plate.getInputMap(JComponent.WHEN_FOCUSED).put(ks, "c-T");
		plate.getActionMap().put("c-T", aTrans);
		plate.setDragEnabled(true);
		plate.setTransferHandler(new TableTransferHandler(annos));
	}
	
	public void stateChanged(ChangeEvent e) {
		lsPlates.setSelectedIndex(tabPlates.getSelectedIndex());
	}

	public void transposePlate() {
		((PlateTableModel)plate.getModel()).transpose();
	}

	public void start() {
    }
	
	public void stop() { //we're not saving on exit
	}

    public void save_plates() {
		//send updated annotations
		//Plates (+ FACSDatas, which are in plates)
		try {
			Vector plates = new Vector();
			Enumeration e = ((DefaultListModel)lsPlates.getModel()).elements();
			while(e.hasMoreElements()) {
				plates.add(e.nextElement());
			}
			//send back annotations for files not in plates, so they're not lost...
			e = ((DefaultListModel)lsFiles.getModel()).elements();
			while(e.hasMoreElements()) {
				plates.add(annos.get(e.nextElement()));
			}
			
			String location = SERVLET_URL;
			location += "?directory="+URLEncoder.encode(getParameter("directory"),"UTF-8")+"&mode="+"send";
			URL servletURL = new URL(getCodeBase(),location);
			URLConnection servlet = servletURL.openConnection();
			servlet.setDoOutput(true);

			servlet.setUseCaches (false); //don't cache connection
			servlet.setDefaultUseCaches (false);

			servlet.setRequestProperty("Content-Type","application/octet-stream");

			ObjectOutputStream oos = new ObjectOutputStream(new BufferedOutputStream(servlet.getOutputStream()));
			oos.writeObject(plates);
			oos.flush();
			oos.close();

			InputStream is = servlet.getInputStream(); //needed to actually send data...
			is.read();

		} catch (IOException ioe) {
			ioe.printStackTrace();
		}
    }

    public void destroy() {
    }
}

class AlphabetListModel extends AbstractListModel implements TableModelListener {
	int rows;

	public AlphabetListModel(int rows) {
		this.rows = rows;
	}

	public int getSize() {
		return rows;
	}

	public Object getElementAt(int index) {
		if(index == 0)
			return "A";
		int a = (int)new Character('A').charValue();
		char[] c = { (char)(a+index) };
		return new String(c);
	}

	public void tableChanged(TableModelEvent e) {
		rows = ((TableModel)e.getSource()).getRowCount();
		fireContentsChanged(this,0,rows);
	}
}

/*class PlateList extends JList implements DropTargetListener {
	//list of plates (DnD capability)
	public Hashtable annos;

	DropTarget dropTarget = new DropTarget (this, this);

	public PlateList(Hashtable annos) {
		this.annos = annos;

		setModel(new DefaultListModel());
		setCellRenderer(new PlateRenderer());
		//setTransferHandler(new ListTransferHandler());
	}

	public void dragEnter (DropTargetDragEvent dropTargetDragEvent) {
		dropTargetDragEvent.acceptDrag (DnDConstants.ACTION_MOVE);
	}

	public void dragExit (DropTargetEvent dropTargetEvent) {}
	public void dragOver (DropTargetDragEvent dropTargetDragEvent) {}
	public void dropActionChanged (DropTargetDragEvent dropTargetDragEvent){}

	public synchronized void drop (DropTargetDropEvent dropTargetDropEvent) {
		try {
			Transferable tr = dropTargetDropEvent.getTransferable();
			if (tr.isDataFlavorSupported(DataFlavor.stringFlavor)) {
				//dropTargetDropEvent.acceptDrop (DnDConstants.ACTION_COPY_OR_MOVE);
				dropTargetDropEvent.acceptDrop (DnDConstants.ACTION_MOVE);
				String fileList = (String)tr.getTransferData(DataFlavor.stringFlavor);
				String[] list = fileList.split("\n");
				int i = locationToIndex(dropTargetDropEvent.getLocation());
				Plate p = (Plate)getModel().getElementAt(i);
				LinkedList ll = getFDsFromName(annos,list);
				p.add(ll);
				dropTargetDropEvent.getDropTargetContext().dropComplete(true);
			} else {
				dropTargetDropEvent.rejectDrop();
			}
		} catch (IOException io) {
			io.printStackTrace();
			dropTargetDropEvent.rejectDrop();
		} catch (UnsupportedFlavorException ufe) {
			ufe.printStackTrace();
			dropTargetDropEvent.rejectDrop();
		}
	}

	public static LinkedList getFDsFromName(Hashtable annos, String[] name) {
		LinkedList ll = new LinkedList();
		for(int n = 0; n < name.length; n++) {
			if(name[n].equals("") || !annos.containsKey(name[n])) {
				//ll.addLast(new Blank(1));
				ll.addLast(new Hashtable());
			} else {
				ll.addLast(annos.get(name[n]));
			}
		}
		return ll;
	}
}*/

class PlateRenderer extends JPanel implements ListCellRenderer {
	//renderer for showing Plate objects in the PlateList
	Plate p;

	public PlateRenderer() {
		p = null;
		setOpaque(true);
	}

	public PlateRenderer(Plate p) {
		this.p = p;
		setLayout(new BoxLayout(this,BoxLayout.Y_AXIS));
		JLabel l1 = new JLabel(p.getRows()+" x "+p.getCols());
		/*l1.setOpaque(true);
		System.out.println(getBackground());
		l1.setBackground(getBackground());
		l1.setForeground(getForeground());*/
		add(l1);
		add(new JLabel(p.getFullName()));
	}

	public java.awt.Component getListCellRendererComponent(JList list,Object value,int index,boolean isSelected,boolean cellHasFocus) {
		if(value instanceof Plate) {
			PlateRenderer pr = new PlateRenderer((Plate)value);
			if (isSelected) {
				pr.setBackground(list.getSelectionBackground());
				pr.setForeground(list.getSelectionForeground());
			} else {
				pr.setBackground(list.getBackground());
				pr.setForeground(list.getForeground());
			}
			return pr;
		} else {
			add(new JLabel("error"));
			return this;
		}

        /*ImageIcon icon = images[selectedIndex];
        String pet = petStrings[selectedIndex];
        setIcon(icon);
        if (icon != null) {
            setText(pet);
            setFont(list.getFont());
        } else {
            setUhOhText(pet + " (no image available)",
                        list.getFont());
        }*/
    }
}

class PropertiesTableModel extends AbstractTableModel {
	private Hashtable fd;
	private AnnotatorClient ac; //just to tell it to update
	private Vector rows;

	public PropertiesTableModel(AnnotatorClient ac) {
		fd = null;
		this.ac = ac;
		rows = new Vector();
	}

	public void setFD(Hashtable fd) {
		this.fd = fd;
		rows = new Vector();
		if(fd != null) {
			Set s = fd.keySet();
			rows.addAll(s);
			Collections.sort(rows);
		}
		fireTableDataChanged();
	}
	
	public Hashtable getFD() {
		return fd;
	}

	public int getColumnCount() {
		return 2;
	}

	public int getRowCount() {
		return rows.size();
	}

	public String getColumnName(int col) {
		return "";
	}

	public String getRowName(int row) {
		return "";
	}

	public Object getValueAt(int row, int col) {
		if(col == 0)
			return rows.elementAt(row);
		return fd.get((String)rows.elementAt(row));
	}

	public Class getColumnClass(int c) {
		return String.class;
	}

	public boolean isCellEditable(int row, int col) {
		/*if(col == 1)
			return true;*/
		return false;
	}

	public void setValueAt(Object value, int row, int col) {
		fd.put((String)rows.elementAt(row),value);
		((PlateTableModel)ac.plate.getModel()).fireTableDataChanged(); //(not always necessary)
	}

	public int getRow(String key) {
		for(int i = 0; i < rows.size(); i++) {
			if(((String)rows.elementAt(i)).equals(key)) {
				return i;
			}
		}
		return -1;
	}

	public String[] getKeys(int[] r) {
		String[] keys = new String[r.length];
		for(int i = 0; i < r.length; i++) {
			keys[i] = (String)rows.elementAt(r[i]);
		}
		return keys;
	}
}

class FileComparator implements Comparator {
	Hashtable annos;
	String key;
	
	public FileComparator(Hashtable annos, String key) {
		this.annos = annos;
		this.key = key;
	}
	
	public int compare(Object o1, Object o2) {
		String s1 = (String)o1;
		String s2 = (String)o2;
		String val1, val2;
		try {
			val1 = (String)((Hashtable)annos.get(s1)).get(key);
			val2 = (String)((Hashtable)annos.get(s2)).get(key);
		} catch (NullPointerException npe) {
			throw new ClassCastException();
		}
		if(val1 == null || val2 == null)
			return 0;
		return val1.compareTo(val2);
	}
	
	public boolean equals(Object obj) {
		if(obj instanceof FileComparator) {
			if(((FileComparator)obj).key.equals(key))
				return true;
		}
		return false;
	}
}

class FileRenderer extends JLabel implements ListCellRenderer {
	Hashtable annos;
	
	String key;
	
	public FileRenderer(Hashtable annos) {
		this.annos = annos;
		key = "";
		
        setOpaque(true);
        setHorizontalAlignment(LEFT);
        setVerticalAlignment(CENTER);
    }
	
	public void setKey(String key) {
		this.key = key;
	}
	
	public String getKey() {
		return key;
	}

    public Component getListCellRendererComponent(
                                       JList list,
                                       Object value,
                                       int index,
                                       boolean isSelected,
                                       boolean cellHasFocus) {
        //Get the selected index. (The index param isn't
        //always valid, so just use the value.)
        //int selectedIndex = ((Integer)value).intValue();

        if (isSelected) {
            setBackground(list.getSelectionBackground());
            setForeground(list.getSelectionForeground());
        } else {
            setBackground(list.getBackground());
            setForeground(list.getForeground());
        }
        
        Hashtable file_annos = (Hashtable)annos.get(value);
        if(!key.equals("") && file_annos.containsKey(key))
        	setText((String)file_annos.get(key));
        else
        	setText((String)value);

        return this;
    }
}

class NullComponent extends JComponent {
	public NullComponent() {
		
	}
	
	public Dimension getPreferredSize() {
		return new Dimension(1,1);
	}
	
	public Dimension getMaximumSize() {
		return getPreferredSize();
	}
	
	public Dimension getMinimumSize() {
		return getPreferredSize();
	}
}