////////////////////////////
// Server side of web-interface experiment annotator
//
// by Mark Hammer
////////////////////////////

package FACS.Annotator;

import FACS.*;

import java.io.*;

import javax.servlet.*;
import javax.servlet.http.*;
import java.util.Vector;
import java.util.Hashtable;
import java.beans.XMLEncoder;
import java.beans.XMLDecoder;
import javax.naming.Context;

public class AnnotatorServlet extends HttpServlet {

	public static final java.beans.ExceptionListener el = new java.beans.ExceptionListener() {
			public void exceptionThrown(Exception exception) {
				exception.printStackTrace();
			}
		};

	public static String ROOT; // = "C:\\Program Files\\Apache Group\\Apache2\\htdocs\\FACS";

	public AnnotatorServlet() {
		super();

		try {
			Context c = new javax.naming.InitialContext();
			ROOT = (String)c.lookup("java:comp/env/fcsroot");
		} catch (javax.naming.NamingException ne) {
			ne.printStackTrace();
			ROOT = "";
		}
	}

	public void doGet(HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException {
	  
		// Use "request" to read incoming HTTP headers (e.g. cookies)
		// and HTML form data (e.g. data the user entered and submitted)

		// Use "response" to specify the HTTP response line and headers
		// (e.g. specifying the content type, setting cookies).

		/*PrintWriter out = response.getWriter();
		// Use "out" to send content to browser
		out.println("AnnotatorServlet");*/
		doPost(request,response);
	}

	public void doPost(HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException {
		String directory = request.getParameter("directory");
		String mode = request.getParameter("mode");
		if(mode.equals("get")) {
			sendAnnos(directory,response);
		} else if(mode.equals("send")) {
			getAnnos(directory,request);
			response.getOutputStream().write(0);
		}
	}

	public void sendAnnos(String directory, HttpServletResponse response) throws ServletException, IOException {
		File dir = new File(ROOT,directory);
		if(!dir.exists() || !dir.isDirectory()) {
			return;
		}

		//send file annotations to client
		//first, read in old annotations
		Hashtable annos = new Hashtable();
		File xml = new File(dir, "annos.xml");
		if(xml.exists()) {
			try {
				XMLDecoder d = new XMLDecoder(new BufferedInputStream(new FileInputStream(xml)));
				d.setExceptionListener(el);
				try {
					annos = (Hashtable)d.readObject(); 
				} catch(ArrayIndexOutOfBoundsException aioobe) { //file is read completely
				} finally {
					d.close();
				}
			} catch (FileNotFoundException fnfe) {
			}
		}
		//now, read in any new FCS files
		File f2;
		File filedir = new File(ROOT,XMLFiles.getFileLink(ROOT,directory));
		String[] d1 = filedir.list();
		for(int i = 0; i < d1.length; i++) {
			if(annos.containsKey(d1[i]))
				continue;
			try {
				f2 = new File(filedir, d1[i]);
				byte[] header = FCSTools.getFCSHeader(f2);
				annos.put(d1[i],FCSTools.readHeader(i,d1[i],header).getDefaultAnnotations());
			} catch (ReadException re) {
				if(!re.toString().startsWith("Not an FCS"))
					System.out.println(re);
			} //wasn't FCS file
		}
		
		//finally, send annos to client
		ObjectOutputStream oos = new ObjectOutputStream(new BufferedOutputStream(response.getOutputStream()));
		try {
			oos.writeObject(annos);
			oos.flush();
		} catch (IOException ioe) {
			ioe.printStackTrace();
			return;
		}

		Vector plates = new Vector();

		//send plate annotations to client
		xml = new File(dir,"plates.xml");
		if(xml.exists()) {
			try {
				XMLDecoder d = new XMLDecoder(new BufferedInputStream(new FileInputStream(xml)));
				d.setExceptionListener(el);
				while(true) {
					try {
						plates.add((Plate)d.readObject());
					} catch(ArrayIndexOutOfBoundsException aioobe) { //file is read completely
						d.close();
						break;
					}
				}
			} catch (FileNotFoundException fnfe) {}
		}

		try {
			oos.writeObject(plates);
			oos.flush();
		} catch (IOException ioe) {
			ioe.printStackTrace();
			return;
		}
		oos.close();
	}

	public void getAnnos(String directory, HttpServletRequest request) throws ServletException, IOException {
		File dir = new File(ROOT,directory);
		File xml = new File(dir,"plates.xml");
		xml.delete();

		Hashtable annos = new Hashtable();
		Vector plates = new Vector();

		ObjectInputStream ois = new ObjectInputStream(new BufferedInputStream(request.getInputStream()));
		try {
			plates = (Vector)ois.readObject();
		} catch (ClassNotFoundException cnfe) {
			cnfe.printStackTrace();
		}

		XMLEncoder xenc = new XMLEncoder(new BufferedOutputStream(new FileOutputStream(xml)));
		xenc.setExceptionListener(el);
		for(int i = 0; i < plates.size(); i++) {
			if(plates.elementAt(i) instanceof Plate) {
				Plate p = (Plate)plates.elementAt(i);
				saveFCS(p,annos);
				xenc.writeObject(p);
				xenc.flush();
			} else { //files not in a plate
				Hashtable fd = (Hashtable)plates.elementAt(i);
				annos.put(fd.get("filename"),fd);
			}
		}
		xenc.close();

		xml = new File(dir,"annos.xml");
		xml.delete();
		xenc = new XMLEncoder(new BufferedOutputStream(new FileOutputStream(xml)));
		xenc.setExceptionListener(el);
		xenc.writeObject(annos);
		xenc.close();

	}

	/*public byte[] getFCSHeader(File f) throws ReadException {
		try {
			FileInputStream fis = new FileInputStream(f);
			byte[] six = new byte[6];
			fis.read(six);
			String s = new String(six);
			if(!s.equals("FCS1.0") && !s.equals("FCS2.0") && !s.equals("FCS3.0"))
				throw new ReadException("Not an FCS File!");
			
			//spaces
			fis.skip(4);
			
			//starts and ends of TEXT, DATA, ANALYSIS
			byte[] eight = new byte[8];
			fis.read(eight);
			int txtStart = getInt(eight);
			fis.read(eight);
			int txtEnd = getInt(eight);

			fis.close();
			fis = new FileInputStream(f);
			byte[] out = new byte[txtEnd];
			fis.read(out);
			fis.close();
			return out;
		} catch (IOException ioe) {
			throw new ReadException(ioe.toString());
		}
	}
    
    public int getInt(byte[] b) throws NumberFormatException {
        String s = new String(b);
        Integer i = new Integer(s.trim());
        return i.intValue();
    }*/

	public void saveFCS(Plate p, Hashtable annos) { //save data annotations to FCS files, and change FACSData->String (for serialization)
		for(int i = 0; i < p.getData().size(); i++) {
			Hashtable fd = (Hashtable)p.getData().elementAt(i);
			if(fd != null && fd.containsKey("filename")) {
				//writeFCS(dir, fd);
				annos.put(fd.get("filename"),fd);
				p.getData().setElementAt(fd.get("filename"),i);
			} else {
				p.getData().setElementAt("",i);
			}
		}
	}
}