///////////////////
//Class that contains a function for interpreting an FCS header

package FACS;

import java.io.*;
import java.util.*;
import java.util.regex.Pattern;
import java.util.regex.Matcher;

public class FCSTools {
	public static byte[] getFCSHeader(File f) throws ReadException {
		try {
            FileInputStream fis = new FileInputStream(f);
			byte[] six = new byte[6];
			fis.read(six);
			String s = new String(six);
			if(!s.equals("FCS1.0") && !s.equals("FCS2.0") && !s.equals("FCS3.0"))
				throw new ReadException("Not an FCS File!");
			
			//spaces
			fis.skip(4);
            
            //starts and ends of TEXT, DATA, ANALYSIS
            byte[] eight = new byte[8];
            fis.read(eight);
            //int txtStart = FCSTools.getInt(eight);
            fis.read(eight);
            int txtEnd = FCSTools.getInt(eight);

			fis.close();
			fis = new FileInputStream(f);
			byte[] out = new byte[txtEnd];
			fis.read(out);
			fis.close();
			return out;
		} catch (IOException ioe) {
			throw new ReadException(ioe.toString());
		}
	}

	public static FACSData readHeader(int dataIndex, String file, byte[] f) throws ReadException {
        try {
            ByteArrayInputStream fis = new ByteArrayInputStream(f);
            
            //file type specified "FCSn.0"
            byte[] six = new byte[6];
            fis.read(six);
            String s = new String(six);
            if(!s.equals("FCS1.0") && !s.equals("FCS2.0") && !s.equals("FCS3.0"))
                throw new ReadException("File Format Unrecognized\n(says "+s+")");
            
            //spaces
            fis.skip(4);
            
            //starts and ends of TEXT, DATA, ANALYSIS
            byte[] eight = new byte[8];
            fis.read(eight);
            int txtStart = getInt(eight);
            fis.read(eight);
            int txtEnd = getInt(eight);
            fis.read(eight);
            int datStart = getInt(eight);
            fis.read(eight);
            int datEnd = getInt(eight);
            //don't care about analysis
            
            if(txtStart == 0 || txtEnd == 0)
                throw new ReadException("FCS3.0 Long Files Unsupported If TEXT Section is long");
            
            String txt; //byte[] data;
            if(datStart == 0) {
            	fis.skip(txtStart - (6+4+8+8+8+8));
                
                byte[] text = new byte[txtEnd-txtStart+1];
                fis.read(text);
                txt = new String(text);
                
                //need to get data start & end from text...
                String del = txt.substring(0,1); //delimiter
                Matcher non = Pattern.compile("[^A-Za-z]").matcher(del);
                String del1 = del;
                if(non.find())
                    del1 = "\\"+del;
                String[] list = txt.split(del1);
                Pattern patBegin = Pattern.compile("BEGINDATA");
                Pattern patEnd = Pattern.compile("ENDDATA");
                for(int i = 0; i < list.length; i++) {
                	if(patBegin.matcher(list[i]).find()) {
                		datStart = new Integer(list[i+1].trim()).intValue();
                	} else if(patEnd.matcher(list[i]).find()) {
                		datEnd = new Integer(list[i+1].trim()).intValue();
                	}
                }
                
                if(datStart == 0 || datEnd == 0)
                	throw new ReadException("Error in FCS3 Long File Data Begin/End Keys");
                
                /*data = new byte[datEnd-datStart+1];
                fis.read(data);*/
            } else if(txtStart < datStart) {
                fis.skip(txtStart - (6+4+8+8+8+8));
                
                byte[] text = new byte[txtEnd-txtStart+1];
                fis.read(text);
                txt = new String(text);
                
                fis.skip(datStart-txtEnd-1);
                
                /*data = new byte[datEnd-datStart+1];
                fis.read(data);*/
            } else {
                fis.skip(datStart - (6+4+8+8+8+8));
                
                /*data = new byte[datEnd-datStart+1];
                fis.read(data);*/
                
                fis.skip(txtStart-datEnd-1);
                
                byte[] text = new byte[txtEnd-txtStart+1];
                fis.read(text);
                txt = new String(text);
            }

            
            Hashtable info = readText(txt);
            
            int mode;
            if(!info.containsKey("MODE"))
                throw new ReadException("File Format Incorrect:\nstorage mode not specified");
            if(info.get("MODE").equals("L")) //list
                mode = 1;
            else if(info.get("MODE").equals("C")) //correlated
                mode = 2;
            else if(info.get("MODE").equals("U")) //uncorr.
                mode = 3;
            else
                throw new ReadException("File Format Incorrect:\nmode unrecognized");
            if(mode != 1)
                throw new ReadException("Only list data supported.");
            
            int nParam;
            if(!info.containsKey("PAR"))
                throw new ReadException("File Format Incorrect:\nparameters unspecified");
            Integer I = new Integer((String)info.get("PAR"));
            nParam = I.intValue();
            
            //set up parameters
            FParameter[] params = new FParameter[nParam];
            int lenSum = 0;
            for(int i = 1; i <= params.length; i++) { //params start at #1
                if(!info.containsKey("P"+i+"B")) //bit length
                    throw new ReadException("File Format Incorrect:\nparameter "+i+" incomplete");
                I = new Integer((String)info.get("P"+i+"B"));
                int pb = I.intValue();
                if(pb % 8 != 0)
                    throw new ReadException("Only byte-justified data supported");
                pb = pb/8;
                lenSum += pb;
                
                if(!info.containsKey("P"+i+"R")) //range
                    throw new ReadException("File Format Incorrect:\nparameter "+i+" incomplete");
                I = new Integer((String)info.get("P"+i+"R"));
                int pe = I.intValue();
                
                if(!info.containsKey("P"+i+"E")) //amplification
                	info.put("P"+i+"E","0,0");
                    //throw new ReadException("File Format Incorrect:\nparameter "+i+" incomplete");
                int comma = ((String)info.get("P"+i+"E")).indexOf(',');
                String Spr1 = ((String)info.get("P"+i+"E")).substring(0,comma);
                String Spr2 = ((String)info.get("P"+i+"E")).substring(comma+1);
                Double d = new Double(Spr1); //was Integer for FCS2.0...
                int pr1 = d.intValue();
                d = new Double(Spr2);
                double pr2 = d.doubleValue();
                
                String ps;
                if(info.containsKey("P"+i+"S"))
                    ps = (String)info.get("P"+i+"S"); //long name
                else
                    ps = "";
                
                String pn;
                
                if(info.containsKey("P"+i+"N"))
                    pn = (String)info.get("P"+i+"N"); //short
                else
                    pn = "P"+i;
                
                params[i-1] = new FParameter(pb, pe, pr1, pr2, ps, pn);
				if(info.containsKey("P"+i+"DISPLAY")) {
					String islog = (String)info.get("P"+i+"DISPLAY");
					if(islog.equalsIgnoreCase("log")) {
						params[i-1].log = true;
						params[i-1].amp1 = (int)Math.ceil(Math.log((double)params[i-1].range)/Math.log(10.0));
					}
				}
            }
            
            //we're going to normalize all parameter ranges to one value
            int paramRange = 0;
            for(int i = 0; i < params.length; i++) {
            	if(params[i].range > paramRange)
            		paramRange = params[i].range;
            }
            for(int i = 0; i < params.length; i++)
            	params[i].range = paramRange;
            
            return new FACSData(dataIndex, file, null, info, params);
        } catch (IOException ioe) {
            throw new ReadException(ioe.toString());
        } catch (NumberFormatException nfe) {
            throw new ReadException("File Corrupted - numbers unreadable: "+nfe.toString());
        }
    }

	public static int getInt(byte[] b) throws NumberFormatException {
        String s = new String(b);
        Integer i = new Integer(s.trim());
        return i.intValue();
    }
    
    public static Hashtable readText(String txt) { //get key-value pairs out of list; keys have "$" at beginning; all are separated by a delimiter
        String del = txt.substring(0,1); //delimiter
        Matcher non = Pattern.compile("[^A-Za-z]").matcher(del);
        String del1 = del;
        if(non.find())
            del1 = "\\"+del;
        String[] list = txt.split(del1);
        
        Hashtable hash = new Hashtable();
        String key = "";
        for(int i = 0; i < list.length; i++) {
            if(i == 0 && list[i].equals(""))
                i++;
            if(list[i].equals("")) {
            	if(i >= list.length)
					break;
                String t = (String)hash.get(key);
                i++;
                hash.put(key,t+del+list[i].trim());
            } else {
            	if(i >= list.length)
					break;
                if(list[i].startsWith("$"))
                    key = list[i].substring(1);
                else
                    key = list[i];
                i++;
                if(i >= list.length) //just if it recognizes something after the last delimiter
                	break;
                hash.put(key,list[i].trim());
            }
        }
        return hash;
    }
}