////////////////////////////
// Client side (applet) of web-interface FACS Plot
//
// by Mark Hammer
////////////////////////////

package FACS.Plot.PlotApplet;

import FACS.FACSData;
import FACS.Plot.Region;
import FACS.Plot.RVector;
import javax.swing.*;
import java.awt.event.*;
import java.awt.BorderLayout;
import java.io.*;
import java.awt.Color;
import java.net.URL;
import java.net.URLConnection;
import java.net.MalformedURLException;
import java.net.URLEncoder;
import java.util.Hashtable;
import java.util.Vector;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import javax.imageio.ImageIO;

import java.awt.image.BufferedImage;

public class PlotApplet extends JApplet implements ActionListener {
	/*public static String HOST;
	public static int port;*/

	public static String PLOTSERVLET_URL, HEADERLET_URL, REGIONLET_URL;
	Hashtable params;

	JPanel p;

	Plot plot;
	/*Socket socket, dSocket;
	PrintWriter out;
	BufferedReader in;*/

	RVector regions;
	Vector data; //"FACSData" descriptors
	String gates; //set by gate() and sent to servlet when asking for a plot
	boolean gatesChanged;
	boolean comp;
	int nEvents;

	boolean saveOnExit;

	PlotKey pk;
	
	public PlotApplet() {
		params = null;
	}
	
	public PlotApplet(Hashtable params) {
		this.params = params;
	}
	
	public String getParameter(String name) {
		if(params == null)
			return super.getParameter(name);
		try {
			String value = (String)params.get(name);
			return value;
		} catch (ClassCastException cce) {
			return null;
		}
	}
	
	public URL getCodeBase() {
		if(params == null)
			return super.getCodeBase();
		return (URL)params.get("codebase");
	}
	
	public void paintPlot(java.awt.Graphics g) {
		plot.paintComponent(g);
	}

	public void init() {
		PLOTSERVLET_URL = "servlets/PlotServlet";
		if(getParameter("plotservlet") != null)
			PLOTSERVLET_URL = getParameter("plotservlet");
		HEADERLET_URL = "servlets/Headerlet";
		if(getParameter("headerlet") != null)
			HEADERLET_URL = getParameter("headerlet");
		REGIONLET_URL = "servlets/Regionlet";
		if(getParameter("regionlet") != null)
			REGIONLET_URL = getParameter("regionlet");

		regions = new RVector();
		data = new Vector();
		gates = "";
		gatesChanged = false;
		saveOnExit = false;

		if(getParameter("comp") != null && getParameter("comp").equals("false"))
			comp = false;
		else
			comp = true;
		
		//number of events to load
		if(getParameter("nEvents") != null)
			nEvents = new Integer(getParameter("nEvents")).intValue();
		else
			nEvents = -1;

		pk = null;

		establishConnection();
		createGUI();
    }

	public void establishConnection() { //talk to server, get files/regions
		try {
			//get regions
			URL servletURL = new URL(getCodeBase(),REGIONLET_URL);
			URLConnection servlet = servletURL.openConnection();
			servlet.setDoInput(true);
			servlet.setDoOutput(true);

			servlet.setUseCaches (false); //don't cache connection
			servlet.setDefaultUseCaches (false);

			PrintStream out = new PrintStream(servlet.getOutputStream());
			out.print("directory="+getParameter("directory"));
			out.close();

			ObjectInputStream ois = new ObjectInputStream(new BufferedInputStream(servlet.getInputStream()));
			try {
				regions = (RVector)ois.readObject();
			} catch (ClassNotFoundException cnfe) {
				cnfe.printStackTrace();
			}
			ois.close();

			//get files (fcs headers)
			servletURL = new URL(getCodeBase(),HEADERLET_URL);

			String[] files = getParameter("file").split(";");
			for(int i = 0; i < files.length; i++) {
				servlet = servletURL.openConnection();
				servlet.setDoInput(true);
				servlet.setDoOutput(true);
				servlet.setUseCaches (false); //don't cache connection
				servlet.setDefaultUseCaches (false);

				out = new PrintStream(servlet.getOutputStream());
				out.print("directory="+getParameter("directory"));
				out.print("&file="+files[i]);
				out.close();

				ois = new ObjectInputStream(new BufferedInputStream(servlet.getInputStream()));
				try {
					FACSData fd = (FACSData)ois.readObject();
					fd.index = i;
					data.addElement(fd);
				} catch (ClassNotFoundException cnfe) {
					cnfe.printStackTrace();
				}
			}

		} catch (java.net.UnknownHostException e) {
            e.printStackTrace();
        } catch (IOException ioe) {
			ioe.printStackTrace();
		}
	}

	public void createGUI() {
		setBackground(Color.white);
		p = new JPanel();
		p.setBackground(Color.white);

		p.setLayout(new BoxLayout(p, BoxLayout.Y_AXIS));

		//toolbar
		JToolBar tb = new JToolBar();
        tb.setFloatable(false);
		JButton bRegions = new JButton();
        URL urlP = PlotApplet.class.getResource("regions.gif");
        if(urlP != null)
            bRegions.setIcon(new ImageIcon(urlP, "Regions..."));
        else
            bRegions.setText("Regions...");
        bRegions.setToolTipText("Regions...");
        bRegions.setActionCommand("regions");
        bRegions.addActionListener(this);
        tb.add(bRegions);
        
        JButton bUnlock = new JButton();
        urlP = PlotApplet.class.getResource("unlock.gif");
        if(urlP != null)
        	bUnlock.setIcon(new ImageIcon(urlP, "Unlock All"));
        else
        	bUnlock.setText("Unlock All");
        bUnlock.setToolTipText("Unlock All");
        bUnlock.setActionCommand("unlock");
        bUnlock.addActionListener(this);
        tb.add(bUnlock);

		JButton bEllipse = new JButton();
        urlP = PlotApplet.class.getResource("ellipse.gif");
        if(urlP != null)
            bEllipse.setIcon(new ImageIcon(urlP, "Ellipse Region"));
        else
            bEllipse.setText("Ellipse Region");
        bEllipse.setToolTipText("Ellipse Region");
        bEllipse.setActionCommand("ellipse");
        bEllipse.addActionListener(this);
        tb.add(bEllipse);
        
        JButton bRect = new JButton();
        urlP = PlotApplet.class.getResource("rect.gif");
        if(urlP != null)
            bRect.setIcon(new ImageIcon(urlP, "Rectangle Region"));
        else
            bRect.setText("Rectangle Region");
        bRect.setToolTipText("Rectangle Region");
        bRect.setActionCommand("rect");
        bRect.addActionListener(this);
        tb.add(bRect);
        
        JButton bPoly = new JButton();
        urlP = PlotApplet.class.getResource("poly.gif");
        if(urlP != null)
            bPoly.setIcon(new ImageIcon(urlP, "Polygon Region"));
        else
            bPoly.setText("Polygon Region");
        bPoly.setToolTipText("Polygon Region");
        bPoly.setActionCommand("poly");
        bPoly.addActionListener(this);
        tb.add(bPoly);
        
        JButton bQuad = new JButton();
        urlP = PlotApplet.class.getResource("quad.gif");
        if(urlP != null)
            bQuad.setIcon(new ImageIcon(urlP, "Quadrants"));
        else
            bQuad.setText("Quadrants");
        bQuad.setToolTipText("Quadrants");
        bQuad.setActionCommand("quad");
        bQuad.addActionListener(this);
        tb.add(bQuad);
        
        JButton bRange = new JButton();
        urlP = PlotApplet.class.getResource("range.gif");
        if(urlP != null)
            bRange.setIcon(new ImageIcon(urlP, "1D Range"));
        else
            bRange.setText("1D Range");
        bRange.setToolTipText("1D Range");
        bRange.setActionCommand("range");
        bRange.addActionListener(this);
        tb.add(bRange);
        
		if(getParameter("compact") == null || getParameter("compact").equals("false"))
	        getContentPane().add(tb, BorderLayout.PAGE_START);


		//plot
		plot = new Plot(this);
		if(params != null)
			plot.setSize(Plot.getPSize());
		else
			p.add(plot);
		
		int p1;
		if(getParameter("param1") != null)
			p1 = new Integer(getParameter("param1")).intValue();
		else
			p1 = -1;
		int p2;
		if(getParameter("param2") != null)
			p2 = new Integer(getParameter("param2")).intValue();
		else
			p2 = -1;
		int m;
		if(getParameter("mode") != null)
			m = new Integer(getParameter("mode")).intValue();
		else
			m = 1;
		if(getParameter("smooth") != null)
			plot.smooth = new Boolean(getParameter("smooth")).booleanValue();

		int[] indices = { 0 };

		Vector gates = new Vector();
		if(getParameter("gates") != null && !getParameter("gates").equals("")) {
			String[] gpcs = getParameter("gates").split(",");
			for(int g = 0; g < gpcs.length; g++) {
				for(int i = 0; i < regions.size(); i++) {
					String rname = ((Region)regions.elementAt(i)).getName();
					if(rname.equals(gpcs[g]))
						gates.add(new Gate(getFDataNames(indices),i,rname));
				}
			}
		}
		plot.gates = gates;
		plot.setPlot(indices, gate(indices,getFDatas(indices),gates), p1, p2, m); //gate(getFDatas(indices),gates)
		if(getParameter("region") != null && !getParameter("region").equals(""))
			plot.setRegion(getParameter("region"));
        //plot.setParams(plot.param1, plot.param2, m);
        
		if(getParameter("region") != null && !getParameter("region").equals(""))
			plot.region = Plot.RECT;

		JButton bOk = new JButton("Save & Close");
		bOk.setActionCommand("okay");
		bOk.addActionListener(this);
		if(getParameter("compact") == null || getParameter("compact").equals("false"))
			p.add(bOk);

		getContentPane().add(p);

		getContentPane().validate();
		plot.regions = plot.translate(regions);
		plot.repaint();
	}

	public void actionPerformed(ActionEvent e) {
		if(e.getActionCommand().equals("okay")) {
			saveOnExit = true;
			updateServerRegions(true);
			try {
				getAppletContext().showDocument(new URL("javascript:closeWindow()"));
			} catch (MalformedURLException murle) {}
		} else if(e.getActionCommand().equals("ellipse")) {
            plot.regionEditing = -1;
            plot.region = Plot.ELLIPSE;
        } else if(e.getActionCommand().equals("rect")) {
            plot.regionEditing = -1;
			plot.region = Plot.RECT;
        } else if(e.getActionCommand().equals("poly")) {
            plot.regionEditing = -1;
			plot.region = Plot.POLY;
        } else if(e.getActionCommand().equals("quad")) {
            plot.regionEditing = -1;
			plot.region = Plot.QUAD;
        } else if(e.getActionCommand().equals("range")) {
            plot.regionEditing = -1;
			plot.region = Plot.RANGE;
        } else if(e.getActionCommand().equals("regions")) {
			RegionsDialog rd = new RegionsDialog(true,plot.fData[0].filename,plot.regions,plot.rDisplayed,plot.rLocked,plot.fData[0].params);
            if(rd.okay) {
                plot.rDisplayed = rd.rDisplayed;
				plot.rLocked = rd.rLocked;
            }
            plot.setVisibilities();
			regions = plot.unTranslateAll(plot.regions);
			updatePlots(0,true);
		} else if(e.getActionCommand().equals("unlock")) {
			plot.unlockAll();
		}
	}

	public String[] getFDataNames(int[] in) {
		FACSData[] fds = getFDatas(in);
		String[] out = new String[in.length];
		for(int i = 0; i < fds.length; i++) {
			out[i] = fds[i].filename;
		}
		return out;
	}

	public void updatePlots(int gateNum, boolean regionExists) {
		//send regions to server (so it can record changes)
		
		boolean newRegion = regions.size() > plot.regions.size();
		if(!regionExists) { //add to list of potentially displayed
			plot.rDisplayed.removeElementAt(gateNum);
			if(plot.regionEditing == gateNum)
				plot.regionEditing = -1;
			else if(plot.regionEditing > gateNum)
				plot.regionEditing--;
		} else {
			if(newRegion) {
				plot.rDisplayed.add(new Boolean(false));
			}
		}
		plot.regions = plot.translate(regions);
		/*if(!regionExists) { //add to list of labels
			plot.percents.removeElementAt(gateNum);
		} else {
			if(newRegion) {
				plot.percents.add(plot.generatePctLabel(plot.regions, gateNum));
			} else {
				plot.percents.setElementAt(plot.generatePctLabel(plot.regions, gateNum),gateNum);
			}
		}*/
		updateServerRegions(false);
	}

	public void updateServerRegions(boolean save) {
		try {
			String location = PLOTSERVLET_URL+"?directory="+URLEncoder.encode(getParameter("directory"),"UTF-8")
				+"&mode=regions"+"&save="+new Boolean(save).toString();
			URL servletURL = new URL(getCodeBase(),location);
			URLConnection servlet = servletURL.openConnection();
			servlet.setDoInput(true);
			servlet.setDoOutput(true);

			servlet.setUseCaches (false); //don't cache connection
			servlet.setDefaultUseCaches (false);

			servlet.setRequestProperty("Content-Type","application/octet-stream");

			BufferedOutputStream os = new BufferedOutputStream(servlet.getOutputStream());
			ObjectOutputStream oos = new ObjectOutputStream(os);
			oos.writeObject(regions);
			oos.flush();
			oos.close();
			os.close();

			InputStream is = servlet.getInputStream();
			is.read();

			/*ByteArrayOutputStream baos = new ByteArrayOutputStream();
			ObjectOutputStream oos = new ObjectOutputStream(baos);
			for(int i = 0; i < regions.size(); i++) { //regions
				Region r = (Region)regions.elementAt(i);
				oos.writeObject(r);
				oos.flush();
			}
			oos.close();
			byte[] b = baos.toByteArray();
			dSocket.getOutputStream().write(b);
			dSocket.getOutputStream().flush();
			out.println("regions "+b.length);*/
		} catch (IOException ioe) {
			ioe.printStackTrace();
		}
	}

	public FACSData[] gate(int[] indices, FACSData[] in, Vector gates) {
		//computation done server-side
		//  server gates its own data... that's what's then used to plot

		StringBuffer b = new StringBuffer();
		//String msg = "gate";
		FACSData[] fds = getFDatas(indices);
		for(int i = 0; i < fds.length; i++) {
			//msg += " "+indices[i];
			if(i != 0)
				b.append("|");
			b.append(fds[i].filename+"");
		}
		//out.println(msg);
		b.append("\n");
		for(int i = 0; i < gates.size(); i++) {
			//out.println(((Gate)gates.elementAt(i)).name+"\t"+((Gate)gates.elementAt(i)).getWritable());
			b.append(((Gate)gates.elementAt(i)).name+"\t"+((Gate)gates.elementAt(i)).getWritable()+"\n");
		}
		//out.println("gate");

		this.gates = b.toString();
		gatesChanged = true;

		return in;
	}

	public FACSData[] getFDatas(int[] indices) {
        FACSData[] f = new FACSData[indices.length];
        for(int i = 0; i < f.length; i++) {
            f[i] = (FACSData)data.elementAt(indices[i]);
        }
        return f;
    }

	public BufferedImage getPlot(int[] indices, int param1, int param2, int mode) {
		try {
			
			URL servletURL = new URL(getCodeBase(),PLOTSERVLET_URL);
			URLConnection servlet = servletURL.openConnection();
			servlet.setDoInput(true);
			servlet.setDoOutput(true);

			servlet.setUseCaches (false); //don't cache connection
			servlet.setDefaultUseCaches (false);

			PrintStream out = new PrintStream(servlet.getOutputStream());
			out.print("directory="+getParameter("directory"));
			out.print("&mode=plot");
			FACSData[] fds = getFDatas(indices);
			String is = fds[0].filename;
			for(int i = 1; i < fds.length; i++) {
				is += ";"+fds[i].filename;
			}
			out.print("&file="+is);
			if(comp)
				out.print("&comp=true");
			else
				out.print("&comp=false");
			out.print("&param1="+param1);
			out.print("&param2="+param2);
			out.print("&pmode="+mode);
			out.print("&hSecondMax="+Boolean.toString(plot.hSecondMax));
			out.print("&outliers="+Boolean.toString(plot.cShowOutliers));
			out.print("&smooth="+Boolean.toString(plot.smooth));
			out.print("&gates="+gates);
			out.print("&gatesChanged="+new Boolean(gatesChanged).toString());
			out.print("&nevents="+nEvents);
			gatesChanged = false;

			out.close();

			BufferedInputStream input = new BufferedInputStream(servlet.getInputStream());
			ByteArrayOutputStream image = new ByteArrayOutputStream(servlet.getContentLength());
			byte[] buf = new byte[servlet.getContentLength()];
			int len = 0;
			while(image.size() < servlet.getContentLength()) {
				len = input.read(buf);
				image.write(buf,0,len);
			}
			ByteArrayInputStream bais = new ByteArrayInputStream(image.toByteArray());
			BufferedImage bi = ImageIO.read(bais);
			
			return bi;
		} catch (IOException ioe) {
			ioe.printStackTrace();
			return new BufferedImage(300,300,BufferedImage.TYPE_4BYTE_ABGR);
		}

		/*String is = ""+indices[0];
		for(int i = 1; i < indices.length; i++) {
			is += ","+indices[i];
		}
		String msg = "plot "+is+" "+param1+" "+param2+" "+mode;
		try {
			out.println(msg);
			String line;
			if(!(line=in.readLine()).startsWith("ready"))
				throw new IOException();
			int len = new Integer(line.substring(6)).intValue();
			int a = 0; byte[] b;
			ByteArrayOutputStream baos = new ByteArrayOutputStream(); //for some reason, Socket InputStream seems to only read up to 8k each time
			while(baos.size() < len) {
				a=dSocket.getInputStream().available();
				b = new byte[a];
				dSocket.getInputStream().read(b);
				baos.write(b);
			}

			BufferedImage bi = ImageIO.read(new ByteArrayInputStream(baos.toByteArray()));
			baos.close();
			return bi;
		} catch (IOException ioe) {
			ioe.printStackTrace();
			return new BufferedImage(300,300,BufferedImage.TYPE_4BYTE_ABGR);
		}*/
	}

	public double gateQuads(int r, int num, FACSData f) { //get percentages for quadrants
		//again, need to do server-side
		return 0;
	}

	public void start() {
    }

    public void stop() {
		//out.println("done");
		//send updated regions
		/*if(saveOnExit)
			updateServerRegions(true);*/

		//close connection
		//out.println("bye");

		/*try {
			in.close();
			out.close();
			socket.close();
			dSocket.close();
		} catch (IOException ioe) {
			ioe.printStackTrace();
		}*/
    }

    public void destroy() {
    }
}

/*class FACSData {
    //store all information contained in a FCS file
    Hashtable text;
    //int[][] data;
    FParameter[] params;
    String name;
    String filename;
    //File f;
    //Vector gates;
    int index;
    
    public FACSData(int index, String filename, String name, Hashtable text, FParameter[] params) { //, int[][] data) {
        this.index = index;
        this.filename = filename;
        this.name = name;
        
        this.text = text;
        //this.data = data;
        this.params = params;
        
        this.name = getName();
    }
    
    public String getName() {
        if(name != null)
            return name;
        if(text.containsKey("SAMPLE ID"))
            return (String)text.get("SAMPLE ID");
        if(text.containsKey("CELLS"))
            return (String)text.get("CELLS");
        if(text.containsKey("COM"))
            return (String)text.get("COM");
        if(text.containsKey("SMNO"))
            return (String)text.get("SMNO");
        if(text.containsKey("SRC"))
            return (String)text.get("SRC");
        if(text.containsKey("PROJ"))
            return (String)text.get("PROJ");
        return filename;
    }
}*/

class ReadException extends Exception {
    String message;
    
    public ReadException(String s) {
        message = s;
    }
    
    public String toString() {
        return message;
    }
}

class FParameter {
    int len, range, amp1;
    double amp2;
    boolean log;
    String lname, sname;
	boolean isLog;
    
    public FParameter(int len, int range, int amp1, double amp2, String lname, String sname) {
        this.len = len;
        this.range = range;
        this.amp1 = amp1;
		if(amp1 == 0)
			isLog = false;
		else
			isLog = true;
        this.amp2 = amp2;
        this.lname = lname;
        this.sname = sname;
        
        if(lname.equals(""))
            this.lname = sname;
        
        if(amp1 == 0 && amp2 == 0)
            log = false;
        else
            log = true;
    }
    
    public int read(byte[] p, int byteorder, int datatype) throws ReadException {
        //in principle, we can read floats and doubles, but I'm lazy
        if(p.length != len)
            throw new ReadException("Data incorrectly formatted.");
        ByteOrder order;
        if(byteorder == 1) {//4,3,2,1
            order = ByteOrder.BIG_ENDIAN;
        } else if(byteorder == 2) { //1,2,3,4
            order = ByteOrder.LITTLE_ENDIAN;
        } else {
            throw new ReadException("3,4,1,2 Byte Order not Supported");
        }
        
        if(datatype == 1) { //integers
            if(len == 2) { //short
                short a = ByteBuffer.wrap(p).order(order).getShort();
				int b = (new Short(a).intValue()) % range;
				if(b < 0)
					b += range;
                return b;
            } else if(len == 4) { //int
                return ByteBuffer.wrap(p).order(order).getInt();
            }
        } else if(datatype == 2) { //float
            if(len == 4) {
                return (int)ByteBuffer.wrap(p).order(order).getFloat();
            }
        } else if(datatype == 3) { //double
            if(len == 8) {
                return (int)ByteBuffer.wrap(p).order(order).getDouble();
            }
        } else { //ASCII
            return new Integer(new String(p)).intValue();
        }
        throw new ReadException("Unreadable data type");
    }
}