/**
 * 
 */
package FACS.Plot.PlotServlet;

import java.util.Enumeration;
import java.util.Vector;
import java.util.concurrent.ConcurrentHashMap;

import javax.servlet.http.HttpSessionBindingEvent;
import javax.servlet.http.HttpSessionBindingListener;

import FACS.FACSData;
import FACS.Plot.RVector;

/**
 * This class will cap the number of elements in the HashMap
 * at the given value, removing the oldest entry if it hits this
 * maximum.
 * @author Mark
 *
 */

public class TimedConcurrentHashMap extends ConcurrentHashMap implements HttpSessionBindingListener {
	private int capacity;
	private static GarbageCollector gc;
	
	public TimedConcurrentHashMap(int capacity) {
		super(capacity);
		this.capacity = capacity;
		if(gc == null) {
			gc = new GarbageCollector(500);
		} else {
			gc.addHashMap();
		}
		if(!gc.isRunning()) {
			gc.startRunning();
			new Thread(gc).start();
		}
	}
	
	public int getCapacity() {
		return capacity;
	}
	
	public Object put(Object key, Object value) {
		TimedObject to = new TimedObject(value);
		if(super.size() == capacity) {
			Enumeration e = super.keys();
			Object removeKey = e.nextElement();
			long removeTime = ((TimedObject)super.get(removeKey)).getTime();
			while(e.hasMoreElements()) {
				Object tkey = e.nextElement();
				long time = ((TimedObject)super.get(tkey)).getTime();
				if(time < removeTime) {
					removeKey = tkey;
					removeTime = time;
				}
			}
			gc.addObject((TimedObject)super.remove(removeKey));
		}
		super.put(key,to);
		return value;
	}
	
	public Object get(Object key) {
		TimedObject to = (TimedObject)super.get(key);
		if(to != null)
			return to.getObject();
		else
			return null;
	}
	
	public void valueBound(HttpSessionBindingEvent hsbe) {
		
	}
	
	public void valueUnbound(HttpSessionBindingEvent hsbe) {
		finalize();
	}
	
	public void finalize() {
		for(Enumeration e = super.elements(); e.hasMoreElements(); ) {
			gc.addObject((TimedObject)super.remove(e.nextElement()));
		}
		gc.removeHashMap();
	}
}

/**
 * This class stores an object and updates the time variable as
 * the object is accessed.
 * 
 * @author Mark
 *
 */
class TimedObject extends Object implements java.io.Serializable {
	private Object o;
	private long time;
	
	public TimedObject(Object o) {
		this.o = o;
		this.time = System.currentTimeMillis();
	}
	
	public Object getObject() {
		time = System.currentTimeMillis();
		return o;
	}
	
	public long getTime() {
		return time;
	}
}

class GarbageCollector implements Runnable, java.io.Serializable {
	private long delay;
	private Vector objects;
	boolean running;
	private int hashmaps;
	
	public GarbageCollector(long delay) {
		this.delay = delay;
		objects = new Vector();
		hashmaps = 0;
	}
	
	public void addHashMap() {
		hashmaps++;
	}
	
	public void removeHashMap() {
		hashmaps--;
	}
	
	public synchronized void startRunning() {
		running = true;
	}
	
	public synchronized boolean isRunning() {
		return running;
	}
	
	public void run() {
		while(running) {
			if(objects.size() == 0 && hashmaps == 0)
				break;
			for(int i = 0; i < objects.size(); i++) {
				TimedObject o = (TimedObject)objects.elementAt(i);
				if(System.currentTimeMillis()-o.getTime() > delay) {
					objects.remove(o);
					i--;
					Object temp = o.getObject();
					if(temp instanceof FACSData) {
						FACSData fd = (FACSData)temp;
						fd.data = null;
						fd = null;
					} else {
						temp = null;
					}
				}
			}
			try {
				System.gc();
				if(objects.size() > 0)
					Thread.sleep(300);
				else
					Thread.sleep(10000);
			} catch (InterruptedException ie) {
			}
		}
		running = false;
	}
	
	public synchronized void stopRunning() {
		running = false;
	}
	
	public synchronized void addObject(TimedObject o) {
		objects.add(o);
		this.notifyAll();
	}
}