//////////////////
//Statistics for a particular population (per file)

package FACS;

import java.util.Vector;
import java.util.Hashtable;

public class PopStats implements java.io.Serializable {
	private static final long serialVersionUID = 1L;
	private String file, pop;
	private double mean[], median[], cv[], pct, num; //arrays are by channel; pct = percent of file in this gate set; num = # cells in this gate
	private Hashtable stats;
	private boolean comp; //was the file compensated before gating?

	private int global; //were the gates global for this file?
	public static final int TRUE = 1;
	public static final int FALSE = 0;

	public PopStats() {
		file = "";
		pop = "";
		mean = new double[0];
		median = new double[0];
		cv = new double[0];
		stats = new Hashtable();
		pct = 0;
		comp = false;
		global = TRUE;
	}

	public PopStats(String file, String pop, boolean comp, boolean global) {
		this.file = file;
		this.pop = pop;
		this.comp = comp;
		if(global)
			this.global = TRUE;
		else
			this.global = FALSE;
		stats = new Hashtable();
	}

	public void setSize(int i) {
		mean = new double[i];
		median = new double[i];
		cv = new double[i];
	}

	public int getSize() {
		return mean.length;
	}

	public void setFile(String file) {
		this.file = file;
	}

	public void setPop(String pop) {
		this.pop = pop;
	}

	public void setComp(boolean comp) {
		this.comp = comp;
	}

	public void setGlobal(int global) {
		this.global = global;
	}

	public void setMean(double[] mean) {
		this.mean = mean;
	}

	public void setMean(int i, double mean) {
		this.mean[i] = mean;
	}

	public void setMedian(double[] median) {
		this.median = median;
	}

	public void setMedian(int i, double median) {
		this.median[i] = median;
	}

	public void setCV(double[] cv) {
		this.cv = cv;
	}

	public void setCV(int i, double cv) {
		this.cv[i] = cv;
	}

	public void setPct(double pct) {
		this.pct = pct;
	}
	
	public void setNum(double num) {
		this.num = num;
	}

	public void setStats(Hashtable stats) {
		this.stats = stats;
	}

	public void setStat(String stat, double[] value) {
		if(stat.equals("Mean"))
			mean = value;
		else if(stat.equals("Median"))
			median = value;
		else if(stat.equals("CV"))
			cv = value;
		else if(stat.equals("Percent"))
			pct = value[0];
		else if(stat.equals("Number"))
			num = value[0];
		else
			stats.put(stat,value);
	}

	public String getFile() {
		return file;
	}

	public String getPop() {
		return pop;
	}

	public double[] getMean() {
		return mean;
	}

	public double[] getMedian() {
		return median;
	}

	public double[] getCV() {
		return cv;
	}

	public double getPct() {
		return pct;
	}
	
	public double getNum() {
		return num;
	}

	public boolean isComp() {
		return comp;
	}

	public int getGlobal() {
		return global;
	}

	public Hashtable getStats() {
		return stats;
	}
	
	public double[] getStat(String stat) {
		if(stat.equals("Mean"))
			return mean;
		else if(stat.equals("Median"))
			return median;
		else if(stat.equals("CV"))
			return cv;
		else if(stat.equals("Percent"))
			return new double[] {pct};
		else if(stat.equals("Number"))
			return new double[] {num};
		else
			return (double[])stats.get(stat);
	}

	public java.util.Enumeration getStatNames() {
		Vector v = new Vector(stats.keySet());
		v.add("Mean");
		v.add("Median");
		v.add("CV");
		v.add("Percent");
		v.add("Number");
		return v.elements();
	}

	public String getHTML() {
		StringBuffer out = new StringBuffer();
		out.append("<table>");

		out.append("<tr>");
		out.append("<td colspan=2><b>"+file+"</b></td><td colspan=2><b>"+pop+"</b></td><td>"+round(pct,1)+"% ("+num+")</td>");
		out.append("</tr>");

		out.append("<tr><td>&nbsp;</td>");
		for(int i = 0; i < mean.length; i++) {
			out.append("<td>P"+(i+1)+"</td>");
		}
		out.append("</tr>");

		out.append("<tr><td>Mean</td>");
		for(int i = 0; i < mean.length; i++) { //for each channel
			out.append("<td>"+round(mean[i],2)+"</td>");
		}
		out.append("</tr>");

		out.append("<tr><td>Median</td>");
		for(int i = 0; i < median.length; i++) { //for each channel
			out.append("<td>"+round(median[i],2)+"</td>");
		}
		out.append("</tr>");

		out.append("<tr><td>CV</td>");
		for(int i = 0; i < cv.length; i++) { //for each channel
			out.append("<td>"+round(cv[i],2)+"</td>");
		}
		out.append("</tr>");

		out.append("</table>");
		return out.toString();
	}

	public String round(double in, int dec) {
		in *= Math.pow(10,dec);
		double out = ((double)((int)in))/Math.pow(10,dec);
		return new Double(out).toString();
	}
}