/**
 * UserTools
 * A utility class for the JSPs to check user auth.
 * @author Mark
 *
 */

package FACS;

import javax.servlet.http.Cookie;

import java.util.Date;
import java.util.Vector;
import java.util.Hashtable;
import java.io.*;

public class UserTools {
	public static final int NO_PERM = 0;
	public static final int VIEW_PERM = 1;
	public static final int EDIT_PERM = 2; //permissions must be in increasing numerical value
	public static final int ADMIN_PERM = 3;
	public static final int MAX_PERM = ADMIN_PERM; //permission for user's own files
	
	private static final String COOKIE_NAME = "User";
	private static final int COOKIE_MAX_AGE = 10;
	
	private static final String EXPTS_FILE = "expts.txt";
	private static final String PASS_FILE = "pass.txt";
	private static final String LOG_FILE = "user_log.txt";
	
	public static final String EVERYONE = "Everyone"; //keyword for "all users"
	
	public static final String SAMPLE_DIR = "Sample"; //directory containing sample experiments
	
	
	public static boolean isLoggedIn(Cookie[] cookies) {
		return getUser(cookies) != null;
	}
	
	public static String getUser(Cookie[] cookies) {
		if(cookies == null)
			return null;
		for(int i = 0; i < cookies.length; i++) {
			if(cookies[i].getName().equals(COOKIE_NAME)) {
				return cookies[i].getValue();
			}
		}
		return null;
	}
	
	public static boolean authenticate(String ip, String ROOT, String user, String pass) {
		log(ip,ROOT,user,pass);
		if(user.equals("anonymous")) {
			return true;
		}
		try {
			BufferedReader br = new BufferedReader(new java.io.FileReader(new File(ROOT,PASS_FILE)));
			String line;
			while((line = br.readLine()) != null) {
				String[] pcs = line.split("\t");
				if(pcs[0].equals(user)) {
					if(pcs[1].equals(pass))
						return true;
					else
						return false;
				}
			}
			br.close();
		} catch (java.io.IOException ioe) {
			ioe.printStackTrace();
		}
		return false;
	}
	
	
	/**
	 * This function logs a user's signin process (successful or not).
	 * 
	 * @param ip Remote IP or host name.
	 * @param ROOT Root directory.
	 * @param user Username.
	 * @param pass Provided password (could be email for anonymous).
	 */
	public static void log(String ip, String ROOT, String user, String pass) {
		try {
			FileWriter fw = new FileWriter(new File(ROOT,LOG_FILE), true);
			fw.write("["+new Date().toString()+"]\t"+ip+"\t"+user+"\t"+pass+"\n");
			fw.flush();
			fw.close();
		} catch (IOException ioe) {
			ioe.printStackTrace();
		}
	}
	
	public static Vector getAllUsers(String ROOT) {
		Vector out = new Vector();
		try {
			BufferedReader br = new BufferedReader(new java.io.FileReader(new File(ROOT,PASS_FILE)));
			String line;
			while((line = br.readLine()) != null) {
				String[] pcs = line.split("\t");
				out.add(pcs[0]);
			}
			br.close();
		} catch (java.io.IOException ioe) {
			ioe.printStackTrace();
		}
		return out;
	}
	
	/**
	 * This function adds a username/password entry to the password file
	 * and creates a data directory for the user.
	 * 
	 * @param ROOT Root directory.
	 * @param user Username.
	 * @param pass Password.
	 * @return <code>true</code> unless user existed OR IO error.
	 */
	public static boolean addUser(String ROOT, String user, String pass) {
		File directory = new File(ROOT,user);
		if(directory.exists())
			return false;
		
		try {
			File passfile = new File(ROOT,PASS_FILE);
			if(passfile.exists()) {
				BufferedReader br = new BufferedReader(new java.io.FileReader(passfile));
				String line;
				while((line = br.readLine()) != null) {
					String[] pcs = line.split("\t");
					if(pcs[0].equals(user))
						return false;
				}
				br.close();
			}
			
			FileWriter writer = new FileWriter(passfile, true);
			writer.write(user+"\t"+pass+"\n");
			writer.flush();
			writer.close();
			
			if(!directory.mkdir())
				return false;
			
			return true;
		} catch (IOException ioe) {
			ioe.printStackTrace();
		}
		return false;
	}
	
	/**
	 * This function removes a username/password entry from the password file.
	 * It does not remove the user's data directory (if it exists).
	 * 
	 * @param ROOT Root directory.
	 * @param user Username.
	 * @return <code>true</code> if user existed AND no IO error.
	 */
	public static boolean deleteUser(String ROOT, String user) {
		Vector lines = new Vector();
		boolean exists = false;
		try {
			File passfile = new File(ROOT,PASS_FILE);
			BufferedReader br = new BufferedReader(new java.io.FileReader(passfile));
			String line;
			while((line = br.readLine()) != null) {
				String[] pcs = line.split("\t");
				if(!pcs[0].equals(user)) {
					lines.add(line);
				} else {
					exists = true;
				}
			}
			br.close();
			
			if(!passfile.delete())
				return false;
			
			PrintWriter writer = new PrintWriter(passfile);
			for(int i = 0; i < lines.size(); i++)
				writer.println((String)lines.elementAt(i));
			writer.flush();
			writer.close();
		} catch (java.io.IOException ioe) {
			ioe.printStackTrace();
			return false;
		}
		return exists;
	}

	/**
	 * Recursive function to delete potentially full directories.
	 * 
	 * @param dir directory to delete.
	 * @return <code>true</code> unless IO error.
	 */
	public static boolean deleteDirectory(File dir) {
		if(dir.exists()) {
			File[] files = dir.listFiles();
			for(int i=0; i<files.length; i++) {
				if(files[i].isDirectory()) { 
					deleteDirectory(files[i]);
				} else {
					files[i].delete();
				}
			}
		}
		return dir.delete();
	}
	
	public static Cookie getCookie(String user) {
		Cookie cookie = new Cookie (COOKIE_NAME, user);
		cookie.setMaxAge(COOKIE_MAX_AGE * 60 * 60);
		return cookie;
	}
	
	public static Cookie getEatenLoginCookie(Cookie[] cookies) {
		if(cookies == null)
			return null;
		for(int i = 0; i < cookies.length; i++) {
			if(cookies[i].getName().equals(COOKIE_NAME)) {
				cookies[i].setMaxAge(0);
				return cookies[i];
			}
		}
		return null;
	}
	
	/**
	 * Tells if user is NOT allowed to upload files.
	 * !! Note: because of the sample file stuff, everyone will have a dir
	 * we just won't allow anonymous users to upload stuff
	 * @param ROOT
	 * @param user
	 * @return <code>true</code> if user is an anonymous user
	 * ("anonymous_XXXX")
	 */
	public static boolean isGhostUser(String ROOT, String user) {
		//a "real" user can upload files, i.e. has a directory
		if(user.startsWith("anonymous_")) //BUT is not a temporary anonymous user ID
			return true;
		/*File userDir = new File(ROOT,user);
		if(userDir.exists() && userDir.isDirectory()) {
			return false;
		}*/
		return false;
	}
	
	/*
	 * return a list of all experiments this user can view
	 */
	public static Hashtable getVisibleExpts(String ROOT, String user) {
		Hashtable out = new Hashtable();
		File userDir = new File(ROOT,user);
		if(userDir.exists() && userDir.isDirectory()) { //user's own files are default
			String[] files = userDir.list();
			Vector v = new Vector();
			for(int i = 0; i < files.length; i++)
				v.add(files[i]);
			out.put(user,v);
		}
		
		//sample experiments
		File sampleDir = new File(ROOT,SAMPLE_DIR);
		if(sampleDir.exists() && sampleDir.isDirectory()) {
			String[] files = sampleDir.list();
			Vector v = new Vector();
			for(int i = 0; i < files.length; i++)
				v.add(files[i]);
			out.put("Sample",v);
		}
		
		//get experiments from other users out of the permissions file
		try {
			BufferedReader br = new BufferedReader(new FileReader(new File(ROOT,EXPTS_FILE)));
			String line;
			while((line = br.readLine()) != null) {
				String[] pcs = line.split("\t");
				if(pcs[0].equals(user)) //happens with "Everyone"
					continue;
				if(pcs[2].contains(user) || pcs[3].contains(user) || pcs[4].contains(user) || pcs[2].contains(EVERYONE) || pcs[3].contains(EVERYONE))
					append(out,pcs[0],pcs[1]);
			}
			br.close();
		} catch (IOException ioe) {
			ioe.printStackTrace();
		}
		
		return out;
	}
	
	/**
	 * @param expt Path string to the experiment directory in question
	 * @return <code>true</code> if <code>expt</code> is in the SAMPLE_DIR directory
	 */
	public static boolean isSampleExperiment(String expt) {
		if(expt.startsWith(SAMPLE_DIR))
			return true;
		return false;
	}
	
	public static int getUserPermission(String expt, String ROOT, String user) {
		String[] pcs = expt.split("/");
		if(pcs[0].equals(user))
			return MAX_PERM;
		
		try {
			BufferedReader br = new BufferedReader(new java.io.FileReader(new File(ROOT,EXPTS_FILE)));
			String line;
			while((line = br.readLine()) != null) {
				String[] linepcs = line.split("\t");
				if(linepcs[0].equals(pcs[0]) && linepcs[1].equals(pcs[1])) {
					//we stipulate that linepcs[2] is view permissions, linepcs[3] is edit, 4 is admin
					if(linepcs[4].contains(user)) //ADMIN cannot extend to everyone
						return ADMIN_PERM;
					if(linepcs[3].contains(user) || linepcs[3].contains(EVERYONE))
						return EDIT_PERM;
					if(linepcs[2].contains(user) || linepcs[2].contains(EVERYONE))
						return VIEW_PERM;
					break;
				}
			}
			br.close();
		} catch (java.io.IOException ioe) {
			ioe.printStackTrace();
		}
		return NO_PERM;
	}
	
	public static Vector getUsers(String ROOT, String expt, int perm) {
		Vector out = new Vector();
		String[] pcs = expt.split("/");
		try {
			BufferedReader br = new BufferedReader(new java.io.FileReader(new File(ROOT,EXPTS_FILE)));
			String line;
			while((line = br.readLine()) != null) {
				String[] linepcs = line.split("\t");
				if(linepcs[0].equals(pcs[0]) && linepcs[1].equals(pcs[1])) {
					//we stipulate that linepcs[2] is view permissions, linepcs[3] is edit, 4 is admin
					String users = "";
					if(perm == ADMIN_PERM)
						users = linepcs[4];
					else if(perm == EDIT_PERM)
						users = linepcs[3];
					else if(perm == VIEW_PERM)
						users = linepcs[2];
					String[] u = users.split(",");
					for(int i = 0; i < u.length; i++)
						out.add(u[i]);
					break;
				}
			}
			br.close();
		} catch (java.io.IOException ioe) {
			ioe.printStackTrace();
		}
		return out;
	}
	
	/*
	 * save the user permissions for a given experiment
	 */
	public static void saveUsers(String ROOT, String expt, String[] view, String[] edit, String[] admin) {
		String[] pcs = expt.split("/");
		Vector lines = new Vector();
		File f = new File(ROOT,EXPTS_FILE);
		boolean found = false;
		try {
			RandomAccessFile raf = new RandomAccessFile(f, "r");
			String line;
			while((line = raf.readLine()) != null) {
				String[] linepcs = line.split("\t");
				if(linepcs[0].equals(pcs[0]) && linepcs[1].equals(pcs[1])) {
					found = true;
					linepcs[2] = nonzero(join(view,","));
					linepcs[3] = nonzero(join(edit,","));
					linepcs[4] = nonzero(join(admin,","));
					line = join(linepcs,"\t");
				}
				lines.add(line);
			}
			raf.close();
		} catch (IOException ioe) {
			ioe.printStackTrace();
		}
		
		if(!found) { //we need to make a new line
			String[] linepcs = new String[5];
			linepcs[0] = pcs[0];
			linepcs[1] = pcs[1];
			linepcs[2] = nonzero(join(view,","));
			linepcs[3] = nonzero(join(edit,","));
			linepcs[4] = nonzero(join(admin,","));
			String line = join(linepcs,"\t");
			lines.add(line);
		}
		
		//ok, rewrite the file
		try {
			f.delete();
			RandomAccessFile raf = new RandomAccessFile(f, "rw");
			for(int i = 0; i < lines.size(); i++) {
				raf.writeBytes(lines.elementAt(i)+"\n");
			}
			raf.close();
		} catch (IOException ioe) {
			ioe.printStackTrace();
		}
	}

	/**
	 * This function appends a string value to the end of the vector that occupies
	 * the given key of the hashtable h (or makes a new vector if one does not exist).
	 * 
	 * @param h Hashtable input
	 * @param key 
	 * @param value to append to the Vector.
	 */
	private static void append(Hashtable h, String key, String value) {
		if(h.containsKey(key)) {
			Vector v = (Vector)h.get(key);
			v.add(value);
		} else {
			Vector v = new Vector();
			v.add(value);
			h.put(key,v);
		}
	}
	
	
	/**
	 * Standard join function (joins an array of strings with the glue).
	 * 
	 * @param pcs Array of pieces.
	 * @param glue String that is stuck between the pieces.
	 * @return The joined string.
	 */
	private static String join(String[] pcs, String glue) {
		StringBuffer b = new StringBuffer();
		for(int i = 0; i < pcs.length; i++) {
			b.append(pcs[i]);
			if(i != pcs.length-1)
				b.append(glue);
		}
		return b.toString();
	}
	
	/**
	 * @param in Input string.
	 * @return Input string if that string is non-empty, otherwise a space character.
	 */
	private static String nonzero(String in) {
		if(in.equals(""))
			return ",";
		else
			return in;
	}
}