/////XMLFiles class
//provides JSP pages with functions to read and write pops.xml, stats.xml, annos.xml, plates.xml, comp.xml
//and getFiles for getting a non-XML directory listing

package FACS;

import java.io.*;
import java.beans.XMLEncoder;
import java.beans.XMLDecoder;
import java.util.Enumeration;
import java.util.Vector;
import java.util.Hashtable;
import FACS.Annotator.Plate;

public class XMLFiles {

	public static final java.beans.ExceptionListener el = new java.beans.ExceptionListener() {
				public void exceptionThrown(Exception exception) {
					exception.printStackTrace();
				}
			};

	public static Hashtable getAnnotations(String ROOT, String dir) {
		File xml = new File(ROOT+dir,"annos.xml");
		if(!xml.exists()) {
			return null;
		}
		try {
			XMLDecoder d = new XMLDecoder(new BufferedInputStream(new FileInputStream(xml)));
			d.setExceptionListener(el);
			Hashtable annos = null;
			try {
				annos = (Hashtable)d.readObject();
			} finally { //file is read completely
				d.close();
			}
			return annos;
		} catch(FileNotFoundException fnfe) {
			return null;
		}
	}

	public static Hashtable getMetrics(String ROOT, String dir) {
		File xml = new File(ROOT+dir,"metrics.xml");
		if(!xml.exists()) {
			return null;
		}
		try {
			XMLDecoder d = new XMLDecoder(new BufferedInputStream(new FileInputStream(xml)));
			d.setExceptionListener(el);
			Hashtable metrics = null;
			try {
				metrics = (Hashtable)d.readObject();
			} finally { //file is read completely
				d.close();
			}
			return metrics;
		} catch(FileNotFoundException fnfe) {
			return null;
		}
	}

	public static Vector getFiles(String ROOT, String dir) {
		Vector out = new Vector();
		File directory = new File(ROOT,getFileLink(ROOT,dir));
		String[] d = directory.list();
		for(int i = 0; i < d.length; i++) {
			if(!d[i].toLowerCase().endsWith("xml"))
				out.add(d[i]);
		}
		return out;
	}

	/**
	 * @param ROOT
	 * @param dir
	 * @return number of channels
	 */
	public static int getParamNumber(String ROOT, String dir) {
		return getParamNumber(getPAnnos(ROOT,dir,false));
	}

	/**
	 * @param ROOT
	 * @param dir
	 * @param encode
	 * @return annotations of first file (to get channels)
	 */
	public static Hashtable getPAnnos(String ROOT, String dir, boolean encode) {
		//encode -> encode HTML-sensitive characters
		File xml = new File(ROOT+dir,"annos.xml");
		if(!xml.exists()) {
			return null;
		}
		Hashtable annos = null;
		try {
			XMLDecoder d = new XMLDecoder(new BufferedInputStream(new FileInputStream(xml)));
			d.setExceptionListener(el);
			try {
				annos = (Hashtable)d.readObject();
			} finally { //file is read completely
				d.close();
			}
		} catch(FileNotFoundException fnfe) {
			return null;
		}
		Hashtable out = (Hashtable)annos.get(annos.keys().nextElement());

		if(encode) {
			Enumeration keys = out.keys();
			while(keys.hasMoreElements()) {
				String key = (String)keys.nextElement();
				String val = (String)out.get(key);
				out.put(key,htmlEncode(val));
			}
		}
		
		return out;
	}

	public static int getParamNumber(Hashtable annos) {
		int i = 0;
		while(annos.containsKey("P"+(i+1))) {
			i++;
		}
		return i;
	}

	public static Vector getParamNames(String ROOT, String dir, boolean encode) {
		Hashtable pannos = getPAnnos(ROOT,dir, encode);
		Vector params = new Vector();
		int i = 0;
		while(pannos.containsKey("P"+(i+1))) {
			String name = (String)pannos.get("P"+(i+1));
			params.add(name);
			i++;
		}
		return params;
	}
	
	public static String htmlEncode(String s) {
		if(s == null)
			return s;
	    StringBuffer out = new StringBuffer();
	    for(int i=0; i<s.length(); i++) {
	        char c = s.charAt(i);
	        if(c > 127 || c=='"' || c=='<' || c=='>') {
	           out.append("&#"+(int)c+";");
	        } else {
	            out.append(c);
	        }
	    }
	    return out.toString();
	}

	public static Vector readPopStats(String ROOT, String dir) {
		File xml = new File(ROOT+dir,"stats.xml");
		Vector stats = new Vector();
		try {
			XMLDecoder d = new XMLDecoder(new BufferedInputStream(new FileInputStream(xml)));
			d.setExceptionListener(el);
			while(true) {
				try {
					stats.add(d.readObject());
				} catch(ArrayIndexOutOfBoundsException aioobe) { //file is read completely
					d.close();
					break;
				}
			}
			return stats;
		} catch (FileNotFoundException fnfe) {
			return null;
		}
	}

	public static PopStats getStatsByPop(Vector stats, String pop, boolean comp) {
		for(int j = 0; j < stats.size(); j++) {
			if(((PopStats)stats.elementAt(j)).getPop().equals(pop) && ((PopStats)stats.elementAt(j)).isComp() == comp)
				return (PopStats)stats.elementAt(j);
		}
		return null;
	}

	public static PopStats getStatsByFilePop(Vector stats, String file, String pop, boolean comp) {
		if(stats == null)
			return null;
		for(int j = 0; j < stats.size(); j++) {
			PopStats ps = (PopStats)stats.elementAt(j);
			if(ps.getFile().equals(file) && ps.getPop().equals(pop) && ps.isComp() == comp)
				return ps;
		}
		return null;
	}

	public static double getStat(PopStats ps, String sel_stat, int ch) {
		double[] vals = null;
		try {
			vals = ps.getStat(sel_stat);
			if(vals == null)
				return Double.NaN;
			return vals[ch];
		} catch (ArrayIndexOutOfBoundsException aioobe) {
			return vals[0];
		}
	}

	public static int getStatLength(Vector stats, String sel_stat) {
		for(int i = 0; i < stats.size(); i++) {
			double[] vals = ((PopStats)stats.elementAt(i)).getStat(sel_stat);
			if(vals == null)
				continue;
			return vals.length;
		}
		return 0;
	}

	public static Vector getAllStats(Vector stats) {
		Vector out = new Vector();
		for(int i = 0; i < stats.size(); i++) {
			for(java.util.Enumeration e = ((PopStats)stats.elementAt(i)).getStatNames(); e.hasMoreElements(); ) {
				String s = (String)e.nextElement();
				if(!out.contains(s))
					out.add(s);
			}
		}
		return out;
	}

	public static void savePops(String ROOT, String dir, Vector total) {
		try {
			File xml = new File(ROOT+dir,"pops.xml");
			xml.delete();
			FileOutputStream fos = new FileOutputStream(xml);
			XMLEncoder xenc = new XMLEncoder(fos);
			xenc.setExceptionListener(el);
			for(int i = 0; i < total.size(); i++) {
				xenc.writeObject(total.elementAt(i));
				xenc.flush();
			}
			xenc.close();
			fos.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * @param ROOT
	 * @param dir
	 * @return true if this experiment has a comp matrix
	 */
	public static boolean isComp(String ROOT, String dir) {
		File xml = new File(ROOT+dir,"comp.xml");
		return xml.exists();
	}

	public static double[][] getComp(String ROOT, String dir) {
		File xml = new File(ROOT+dir,"comp.xml");
		double[][] c = null;
		if(xml.exists()) {
			try {
				XMLDecoder d = new XMLDecoder(new BufferedInputStream(new FileInputStream(xml)));
				d.setExceptionListener(el);
				try {
					c = (double[][])d.readObject();
				} finally { //file is read completely
					d.close();
				}
			} catch (IOException ioe) {
				ioe.printStackTrace();
			}
		} else {
			//make a blank comp matrix
			Hashtable annos = getPAnnos(ROOT,dir,true);
			int nParams = getParamNumber(annos);
			c = new double[nParams][nParams];
			for(int i = 0; i < nParams; i++) {
				c[i][i] = 1.0;
			}
		}
		return c;
	}

	public static void saveComp(String ROOT, String dir, double[][] c) {
		try {
			File xml = new File(ROOT+dir,"comp.xml");
			xml.delete();
			FileOutputStream fos = new FileOutputStream(xml);
			XMLEncoder xenc = new XMLEncoder(fos);
			xenc.setExceptionListener(el);
			xenc.writeObject(c);
			xenc.flush();
			xenc.close();
			fos.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	public static Vector getPopulations(String ROOT, String dir) {
		File xml = new File(ROOT+dir,"pops.xml");
		if(!xml.exists()) {
			return new Vector();
		}
		try {
			XMLDecoder d = new XMLDecoder(new BufferedInputStream(new FileInputStream(xml)));
			d.setExceptionListener(el);
			Vector pops = new Vector();
			while(true) {
				try {
					pops.add((Population)d.readObject());
				} catch(ArrayIndexOutOfBoundsException aioobe) { //file is read completely
					d.close();
					break;
				}
			}
			return pops;
		} catch(FileNotFoundException fnfe) {
			return new Vector();
		}
	}

	public static Vector getPopulationsComp(String ROOT, String dir, boolean comp) {
		File xml = new File(ROOT+dir,"pops.xml");
		if(!xml.exists()) {
			return new Vector();
		}
		try {
			XMLDecoder d = new XMLDecoder(new BufferedInputStream(new FileInputStream(xml)));
			d.setExceptionListener(el);
			Vector pops = new Vector();
			while(true) {
				try {
					Population p = (Population)d.readObject();
					if(p.getName().matches("^p\\d{0,1}[PN]$") == comp) {
						pops.add(p);
					}
				} catch(ArrayIndexOutOfBoundsException aioobe) { //file is read completely
					d.close();
					break;
				}
			}
			return pops;
		} catch(FileNotFoundException fnfe) {
			return new Vector();
		}
	}

	public static Population getPopulation(Vector pops, String pname) {
		for(int i = 0; i < pops.size(); i++) {
			if(((Population)pops.elementAt(i)).getName().equals(pname))
				return (Population)pops.elementAt(i);
		}
		return null;
	}

	public static Vector getPopsByGate(Vector pops, String gname) {
		Vector out = new Vector();
		for(int i = 0; i < pops.size(); i++) {
			Vector gates = ((Population)pops.elementAt(i)).getGates();
			for(int j = 0; j < gates.size(); j++) {
				if(((String)gates.elementAt(j)).equals(gname)) {
					out.add((Population)pops.elementAt(i));
					break;
				}
			}
		}
		return out;
	}

	public static String getFileByPop(Vector pops, String pop) { //for populations w/ only one file (returns 1st file)
		for(int i = 0; i < pops.size(); i++) {
			if(((Population)pops.elementAt(i)).getName().equals(pop))
				return (String)((Population)pops.elementAt(i)).getFiles().elementAt(0);
		}
		return null;
	}


	public static Vector getPlates(String ROOT, String dir) {
		File xml = new File(ROOT+dir,"plates.xml");
		try {
			XMLDecoder d = new XMLDecoder(new BufferedInputStream(new FileInputStream(xml)));
			d.setExceptionListener(el);
			Vector out = new Vector();
			while(true) {
				try {
					out.add((Plate)d.readObject());
				} catch(ArrayIndexOutOfBoundsException aioobe) { //file is read completely
					d.close();
					break;
				}
			}
			return out;
		} catch(FileNotFoundException fnfe) {
			return null;
		}
	}

	public static Plate getPlate(Vector plates, String pname) {
		for(int i = 0; i < plates.size(); i++) {
			if(((Plate)plates.elementAt(i)).getFullName().equals(pname))
				return (Plate)plates.elementAt(i);
		}
		return null;
	}
	
	public static String getFileLink(String ROOT, String dir) {
		File xml = new File(ROOT+dir,"link.xml");
		try {
			XMLDecoder d = new XMLDecoder(new BufferedInputStream(new FileInputStream(xml)));
			d.setExceptionListener(el);
			String out = dir;
			try {
				out = (String)d.readObject();
			} finally {
				d.close();
			}
			return out;
		} catch(FileNotFoundException fnfe) {
			return dir;
		}
	}
}